/*
Copyright (C) 2005 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#if _MSC_VER > 1000
#pragma once
#endif

#ifndef ASSEMBLYWRITERPSP_H_
#define ASSEMBLYWRITERPSP_H_

#include "AssemblyBuffer.h"
#include "DynarecTargetPSP.h"

#include "Core/R4300OpCode.h"

class CAssemblyWriterPSP
{
	public:
		CAssemblyWriterPSP( CAssemblyBuffer * p_buffer )
			:	mpAssemblyBuffer( p_buffer )
		{
		}

	public:
		CAssemblyBuffer *	GetAssemblyBuffer() const									{ return mpAssemblyBuffer; }
		void				SetAssemblyBuffer( CAssemblyBuffer * p_buffer )				{ mpAssemblyBuffer = p_buffer; }

	// XXXX
	private:
	public:
		void				LoadConstant( EPspReg reg, s32 value );

		void				LoadRegister( EPspReg reg_dst, OpCodeValue load_op, EPspReg reg_base, s16 offset );
		void				StoreRegister( EPspReg reg_src, OpCodeValue store_op, EPspReg reg_base, s16 offset );

		void				NOP();
		void				LUI( EPspReg reg, u16 value );

		void				SW( EPspReg reg_src, EPspReg reg_base, s16 offset );
		void				SH( EPspReg reg_src, EPspReg reg_base, s16 offset );
		void				SB( EPspReg reg_src, EPspReg reg_base, s16 offset );

		void				LW( EPspReg reg_dst, EPspReg reg_base, s16 offset );

		void				LWC1( EPspFloatReg reg_dst, EPspReg reg_base, s16 offset );
		void				SWC1( EPspFloatReg reg_src, EPspReg reg_base, s16 offset );

		CJumpLocation		JAL( CCodeLabel target, bool insert_delay );
		CJumpLocation		J( CCodeLabel target, bool insert_delay );

		void 				JR( EPspReg reg_link, bool insert_delay );

		CJumpLocation		BNE( EPspReg a, EPspReg b, CCodeLabel target, bool insert_delay );
		CJumpLocation		BEQ( EPspReg a, EPspReg b, CCodeLabel target, bool insert_delay );
		CJumpLocation		BLEZ( EPspReg a, CCodeLabel target, bool insert_delay );
		CJumpLocation		BGTZ( EPspReg a, CCodeLabel target, bool insert_delay );

		CJumpLocation		BLTZ( EPspReg a, CCodeLabel target, bool insert_delay );
		CJumpLocation		BGEZ( EPspReg a, CCodeLabel target, bool insert_delay );
		CJumpLocation		BLTZL( EPspReg a, CCodeLabel target, bool insert_delay );
		CJumpLocation		BGEZL( EPspReg a, CCodeLabel target, bool insert_delay );

		void				ADDI( EPspReg reg_dst, EPspReg reg_src, s16 value );
		void				ADDIU( EPspReg reg_dst, EPspReg reg_src, s16 value );
		void				SLTI( EPspReg reg_dst, EPspReg reg_src, s16 value );
		void				SLTIU( EPspReg reg_dst, EPspReg reg_src, s16 value );

		void				ANDI( EPspReg reg_dst, EPspReg reg_src, u16 immediate );
		void				ORI( EPspReg reg_dst, EPspReg reg_src, u16 immediate );
		void				XORI( EPspReg reg_dst, EPspReg reg_src, u16 immediate );

		void				SLL( EPspReg reg_dst, EPspReg reg_src, u32 shift );
		void				SRL( EPspReg reg_dst, EPspReg reg_src, u32 shift );
		void				SRA( EPspReg reg_dst, EPspReg reg_src, u32 shift );
		void				SLLV( EPspReg rd, EPspReg rs, EPspReg rt );
		void				SRLV( EPspReg rd, EPspReg rs, EPspReg rt );
		void				SRAV( EPspReg rd, EPspReg rs, EPspReg rt );

		void				ADD( EPspReg rd, EPspReg rs, EPspReg rt );
		void				ADDU( EPspReg rd, EPspReg rs, EPspReg rt );
		void				SUB( EPspReg rd, EPspReg rs, EPspReg rt );
		void				SUBU( EPspReg rd, EPspReg rs, EPspReg rt );
		void				AND( EPspReg rd, EPspReg rs, EPspReg rt );
		void				OR( EPspReg rd, EPspReg rs, EPspReg rt );
		void				XOR( EPspReg rd, EPspReg rs, EPspReg rt );
		void				NOR( EPspReg rd, EPspReg rs, EPspReg rt );

		void				SLT( EPspReg rd, EPspReg rs, EPspReg rt );
		void				SLTU( EPspReg rd, EPspReg rs, EPspReg rt );

		void				Cop1Op( ECop1Op cop1_op, EPspFloatReg fd, EPspFloatReg fs, ECop1OpFunction cop1_funct, EPspFloatReg ft );
		void				Cop1Op( ECop1Op cop1_op, EPspFloatReg fd, EPspFloatReg fs, ECop1OpFunction cop1_funct );

		void				MFC1( EPspReg rt, EPspFloatReg fs );
		void				MTC1( EPspFloatReg fs, EPspReg rt );

		void				ADD_S( EPspFloatReg fd, EPspFloatReg fs, EPspFloatReg ft );
		void				SUB_S( EPspFloatReg fd, EPspFloatReg fs, EPspFloatReg ft );
		void				MUL_S( EPspFloatReg fd, EPspFloatReg fs, EPspFloatReg ft );
		void				DIV_S( EPspFloatReg fd, EPspFloatReg fs, EPspFloatReg ft );

		void				SQRT_S( EPspFloatReg fd, EPspFloatReg fs );
		void				ABS_S( EPspFloatReg fd, EPspFloatReg fs );
		void				MOV_S( EPspFloatReg fd, EPspFloatReg fs );
		void				NEG_S( EPspFloatReg fd, EPspFloatReg fs );

		static	void		GetLoadConstantOps( EPspReg reg, s32 value, PspOpCode * p_op1, PspOpCode * p_op2 );

		inline void AppendOp( PspOpCode op )
		{
			mpAssemblyBuffer->EmitDWORD( op._u32 );
		}

	private:
		CJumpLocation		BranchOp( EPspReg a, OpCodeValue op, EPspReg b, CCodeLabel target, bool insert_delay );
		CJumpLocation		BranchRegImmOp( EPspReg a, ERegImmOp op, CCodeLabel target, bool insert_delay );
		void				SpecOpLogical( EPspReg rd, EPspReg rs, ESpecOp op, EPspReg rt );

	private:
		CAssemblyBuffer *				mpAssemblyBuffer;
};

#endif // ASSEMBLYWRITERPSP_H_
