/*
Copyright (C) 2001,2005 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#if _MSC_VER > 1000
#pragma once
#endif

#ifndef CODEGENERATOR_H_
#define CODEGENERATOR_H_

struct	OpCode;
struct	SRegisterStats;
struct	SBranchDetails;
class	CDynarecCode;
class	CIndirectExitMap;

#include "Core/R4300Instruction.h"
#include "AssemblyUtils.h"
#include "RegisterSpan.h"

// XXXX For GenerateCompare_S/D
#define FLAG_SWAP			0x100
#define FLAG_C_LT		(0x41)					// jne-   le
#define FLAG_C_LE		(FLAG_SWAP|0x41)		// je -   ! gt
#define FLAG_C_EQ		(FLAG_SWAP|0x40)		// je -   ! eq


// 
//	Allow an architecure-independent way of passing around register sets
//
struct RegisterSnapshotHandle
{
	explicit RegisterSnapshotHandle( u32 handle ) : Handle( handle ) {}

	u32		Handle;
};


class CCodeGenerator
{
	public:
		typedef void (__fastcall * ExceptionHandlerFn )();

									CCodeGenerator() {}
		virtual						~CCodeGenerator() {}

		virtual void				Finalise() = 0;

		virtual	void				AnalyseStats( const SRegisterStats & stats ) = 0;

		virtual void				SetCachedBasePointer( const void * p_base ) = 0;
		virtual void				SetRegisterSpanList( const RegisterSpanList & span_list ) = 0;
		virtual void				UpdateRegisterCaching( u32 instruction_idx ) = 0;

		virtual RegisterSnapshotHandle	GetRegisterSnapshot() = 0;

		virtual CCodeLabel			GetEntryPoint() const = 0;
		virtual u32					GetCompiledCodeSize() const = 0;

		virtual	void				GeneratePrologue() = 0;
		virtual	CCodeLabel			GenerateEpilogue( u32 pc ) = 0;

		virtual	CJumpLocation		GenerateExitCode( u32 exit_address, u32 jump_address, u32 num_instructions, CCodeLabel next_fragment ) = 0;
		virtual void				GenerateEretExitCode( u32 num_instructions, CIndirectExitMap * p_map ) = 0;
		virtual void				GenerateIndirectExitCode( u32 num_instructions, CIndirectExitMap * p_map ) = 0;
		virtual void				GenerateExceptionHander( ExceptionHandlerFn p_exception_handler_fn, std::vector< CJumpLocation > & exception_handler_jumps ) = 0;

		virtual	void				SetVar( u32 * p_var, u32 value ) = 0;

		virtual	void				CleanupForExit() = 0;

		virtual void				PreEmitCachedRegCheck( const SRegisterStats & stats ) = 0;
		virtual void				PostEmitCachedRegCheck( const SRegisterStats & stats, bool optimised ) = 0;

		virtual void				GenerateBlockEndBranch( CDynarecCode * p_target, u32 num_ops ) = 0;
		virtual	void				GenerateBlockEnd() = 0;

		virtual void				GenerateBranchHandler( CJumpLocation branch_handler_jump, RegisterSnapshotHandle snapshot ) = 0;

		virtual CJumpLocation		GenerateOpCode( u32 address, bool branch_delay_slot, OpCode op_code, const SBranchDetails * p_branch, CJumpLocation * p_branch_jump ) = 0;

		virtual	void				GenerateGenericR4300( bool optimise, OpCode op_code, CPU_Instruction pF ) = 0;

		virtual	void				GenerateMFHI( EN64Reg rd ) = 0;
		virtual	void				GenerateMFLO( EN64Reg rd ) = 0;

		virtual	void				GenerateMULTU( EN64Reg rs, EN64Reg rt ) = 0;
		virtual	void				GenerateADDU( EN64Reg rd, EN64Reg rs, EN64Reg rt ) = 0;
		virtual	void				GenerateSUBU( EN64Reg rd, EN64Reg rs, EN64Reg rt ) = 0;
		virtual	void				GenerateAND( EN64Reg rd, EN64Reg rs, EN64Reg rt ) = 0;
		virtual	void				GenerateOR( EN64Reg rd, EN64Reg rs, EN64Reg rt ) = 0;
		virtual	void				GenerateXOR( EN64Reg rd, EN64Reg rs, EN64Reg rt ) = 0;
		virtual	void				GenerateNOR( EN64Reg rd, EN64Reg rs, EN64Reg rt ) = 0;
		virtual	void				GenerateSLT( EN64Reg rd, EN64Reg rs, EN64Reg rt ) = 0;
		virtual	void				GenerateSLTU( EN64Reg rd, EN64Reg rs, EN64Reg rt ) = 0;

		virtual void				GenerateADDIU( EN64Reg rt, EN64Reg rs, s16 immediate ) = 0;
		virtual void				GenerateANDI( EN64Reg rt, EN64Reg rs, u16 immediate ) = 0;
		virtual void				GenerateORI( EN64Reg rt, EN64Reg rs, u16 immediate ) = 0;
		virtual void				GenerateXORI( EN64Reg rt, EN64Reg rs, u16 immediate ) = 0;
		virtual void				GenerateLUI( EN64Reg rt, s16 immediate ) = 0;
		virtual void				GenerateSLTI( EN64Reg rt, EN64Reg rs, s16 immediate ) = 0;
		virtual void				GenerateSLTIU( EN64Reg rt, EN64Reg rs, s16 immediate ) = 0;
		virtual void				GenerateSLL( EN64Reg rd, EN64Reg rt, u32 sa ) = 0;
		virtual void				GenerateSRL( EN64Reg rd, EN64Reg rt, u32 sa ) = 0;
		virtual void				GenerateSRA( EN64Reg rd, EN64Reg rt, u32 sa ) = 0;

		virtual void				GenerateLB( EN64Reg rt, EN64Reg base, s32 offset ) = 0;
		virtual void				GenerateLBU( EN64Reg rt, EN64Reg base, s32 offset ) = 0;
		virtual void				GenerateLH( EN64Reg rt, EN64Reg base, s32 offset ) = 0;
		virtual void				GenerateLHU( EN64Reg rt, EN64Reg base, s32 offset ) = 0;
		virtual void				GenerateLW( EN64Reg rt, EN64Reg base, s16 offset ) = 0;
		virtual	void				GenerateLWC1( u32 ft, EN64Reg base, s32 offset ) = 0;

		virtual	void				GenerateSB( EN64Reg rt, EN64Reg base, s32 offset ) = 0;
		virtual	void				GenerateSH( EN64Reg rt, EN64Reg base, s32 offset ) = 0;
		virtual	void				GenerateSW( EN64Reg rt, EN64Reg base, s32 offset ) = 0;
		virtual	void				GenerateSWC1( u32 ft, EN64Reg base, s32 offset ) = 0;

		virtual	void				GenerateMFC1( EN64Reg rt, u32 fs ) = 0;
		virtual	void				GenerateMTC1( u32 fs, EN64Reg rt ) = 0;
		virtual	void				GenerateCFC1( EN64Reg rt, u32 fs ) = 0;
		virtual	void				GenerateCTC1( u32 fs, EN64Reg rt ) = 0;

		virtual	void				GenerateBEQ( EN64Reg rs, EN64Reg rt, u32 target ) = 0;
		virtual	void				GenerateBNE( EN64Reg rs, EN64Reg rt, u32 target ) = 0;
		virtual	void				GenerateBLEZ( EN64Reg rs, u32 target ) = 0;
		virtual	void				GenerateBGTZ( EN64Reg rs, u32 target ) = 0;

		virtual	void				GenerateADD_S( u32 fd, u32 fs, u32 ft ) = 0;
		virtual	void				GenerateSUB_S( u32 fd, u32 fs, u32 ft ) = 0;
		virtual	void				GenerateMUL_S( u32 fd, u32 fs, u32 ft ) = 0;
		virtual	void				GenerateDIV_S( u32 fd, u32 fs, u32 ft ) = 0;
		virtual	void				GenerateSQRT_S( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateMOV_S( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateNEG_S( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateCompare_S( u32 fs, u32 ft, u16 flag ) = 0;

		virtual	void				GenerateADD_D( u32 fd, u32 fs, u32 ft ) = 0;
		virtual	void				GenerateSUB_D( u32 fd, u32 fs, u32 ft ) = 0;
		virtual	void				GenerateMUL_D( u32 fd, u32 fs, u32 ft ) = 0;
		virtual	void				GenerateDIV_D( u32 fd, u32 fs, u32 ft ) = 0;
		virtual	void				GenerateSQRT_D( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateMOV_D( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateNEG_D( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateCompare_D( u32 fs, u32 ft, u16 flag ) = 0;

		virtual	void				GenerateCVT_S_W( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateCVT_S_D( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateCVT_D_W( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateCVT_D_S( u32 fd, u32 fs ) = 0;
		virtual	void				GenerateCVT_W_S( u32 fd, u32 fs ) = 0;
};

#endif // CODEGENERATOR_H_
