/*
Copyright (C) 2001,2005 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#if _MSC_VER > 1000
#pragma once
#endif

#ifndef CODEGENERATORX86_H_
#define CODEGENERATORX86_H_

#include "CodeGenerator.h"
#include "AssemblyWriterX86.h"
#include "DynarecTargetX86.h"

class CCodeGeneratorX86 : public CCodeGenerator, public CAssemblyWriterX86
{
	public:
		CCodeGeneratorX86( CAssemblyBuffer * p_primary, CAssemblyBuffer * p_secondary );

		virtual void				Finalise();

		virtual	void				AnalyseStats( const SRegisterStats & stats );

		virtual void				SetCachedBasePointer( const void * p_base );
		virtual void				SetRegisterSpanList( const RegisterSpanList & span_list );
		virtual void				UpdateRegisterCaching( u32 instruction_idx );

		virtual RegisterSnapshotHandle	GetRegisterSnapshot();

		virtual CCodeLabel			GetEntryPoint() const;
		virtual u32					GetCompiledCodeSize() const;

		virtual	void				GeneratePrologue();
		virtual	CCodeLabel			GenerateEpilogue( u32 pc );

		virtual	CJumpLocation		GenerateExitCode( u32 exit_address, u32 jump_address, u32 num_instructions, CCodeLabel next_fragment );
		virtual void				GenerateEretExitCode( u32 num_instructions, CIndirectExitMap * p_map );
		virtual void				GenerateIndirectExitCode( u32 num_instructions, CIndirectExitMap * p_map );
		virtual void				GenerateExceptionHander( ExceptionHandlerFn p_exception_handler_fn, std::vector< CJumpLocation > & exception_handler_jumps );

		virtual	void				SetVar( u32 * p_var, u32 value );
				void				SetVar8( u32 * p_var, u8 value );

		virtual	void				CleanupForExit();

		virtual void				PreEmitCachedRegCheck( const SRegisterStats & stats );
		virtual void				PostEmitCachedRegCheck( const SRegisterStats & stats, bool optimised );

		virtual void				GenerateBlockEndBranch( CDynarecCode * p_target, u32 num_ops );
		virtual	void				GenerateBlockEnd();

		virtual void				GenerateBranchHandler( CJumpLocation branch_handler_jump, RegisterSnapshotHandle snapshot );

		virtual CJumpLocation		GenerateOpCode( u32 address, bool branch_delay_slot, OpCode op_code, const SBranchDetails * p_branch, CJumpLocation * p_branch_jump );

		virtual	void				GenerateGenericR4300( bool optimise, OpCode op_code, CPU_Instruction pF );

		virtual	void				GenerateMFHI( EN64Reg rd );
		virtual	void				GenerateMFLO( EN64Reg rd );

		virtual	void				GenerateMULTU( EN64Reg rs, EN64Reg rt );
		virtual	void				GenerateADDU( EN64Reg rd, EN64Reg rs, EN64Reg rt );
		virtual	void				GenerateSUBU( EN64Reg rd, EN64Reg rs, EN64Reg rt );
		virtual	void				GenerateAND( EN64Reg rd, EN64Reg rs, EN64Reg rt );
		virtual	void				GenerateOR( EN64Reg rd, EN64Reg rs, EN64Reg rt );
		virtual	void				GenerateXOR( EN64Reg rd, EN64Reg rs, EN64Reg rt );
		virtual	void				GenerateNOR( EN64Reg rd, EN64Reg rs, EN64Reg rt );
		virtual	void				GenerateSLT( EN64Reg rd, EN64Reg rs, EN64Reg rt );
		virtual	void				GenerateSLTU( EN64Reg rd, EN64Reg rs, EN64Reg rt );

		virtual void				GenerateADDIU( EN64Reg rt, EN64Reg rs, s16 immediate );
		virtual void				GenerateANDI( EN64Reg rt, EN64Reg rs, u16 immediate );
		virtual void				GenerateORI( EN64Reg rt, EN64Reg rs, u16 immediate );
		virtual void				GenerateXORI( EN64Reg rt, EN64Reg rs, u16 immediate );
		virtual void				GenerateLUI( EN64Reg rt, s16 immediate );
		virtual void				GenerateSLTI( EN64Reg rt, EN64Reg rs, s16 immediate );
		virtual void				GenerateSLTIU( EN64Reg rt, EN64Reg rs, s16 immediate );
		virtual void				GenerateSLL( EN64Reg rd, EN64Reg rt, u32 sa );
		virtual void				GenerateSRL( EN64Reg rd, EN64Reg rt, u32 sa );
		virtual void				GenerateSRA( EN64Reg rd, EN64Reg rt, u32 sa );

		virtual void				GenerateLB( EN64Reg rt, EN64Reg base, s32 offset );
		virtual void				GenerateLBU( EN64Reg rt, EN64Reg base, s32 offset );
		virtual void				GenerateLH( EN64Reg rt, EN64Reg base, s32 offset );
		virtual void				GenerateLHU( EN64Reg rt, EN64Reg base, s32 offset );
		virtual void				GenerateLW( EN64Reg rt, EN64Reg base, s16 offset );
		virtual	void				GenerateLWC1( u32 ft, EN64Reg base, s32 offset );

		virtual	void				GenerateSB( EN64Reg rt, EN64Reg base, s32 offset );
		virtual	void				GenerateSH( EN64Reg rt, EN64Reg base, s32 offset );
		virtual	void				GenerateSW( EN64Reg rt, EN64Reg base, s32 offset );
		virtual	void				GenerateSWC1( u32 ft, EN64Reg base, s32 offset );

		virtual	void				GenerateMFC1( EN64Reg rt, u32 fs );
		virtual	void				GenerateMTC1( u32 fs, EN64Reg rt );
		virtual	void				GenerateCFC1( EN64Reg rt, u32 fs );
		virtual	void				GenerateCTC1( u32 fs, EN64Reg rt );

		virtual	void				GenerateBEQ( EN64Reg rs, EN64Reg rt, u32 target );
		virtual	void				GenerateBNE( EN64Reg rs, EN64Reg rt, u32 target );
		virtual	void				GenerateBLEZ( EN64Reg rs, u32 target );
		virtual	void				GenerateBGTZ( EN64Reg rs, u32 target );

		virtual	void				GenerateADD_S( u32 fd, u32 fs, u32 ft );
		virtual	void				GenerateSUB_S( u32 fd, u32 fs, u32 ft );
		virtual	void				GenerateMUL_S( u32 fd, u32 fs, u32 ft );
		virtual	void				GenerateDIV_S( u32 fd, u32 fs, u32 ft );
		virtual	void				GenerateSQRT_S( u32 fd, u32 fs );
		virtual	void				GenerateMOV_S( u32 fd, u32 fs );
		virtual	void				GenerateNEG_S( u32 fd, u32 fs );
		virtual	void				GenerateCompare_S( u32 fs, u32 ft, u16 flag );

		virtual	void				GenerateADD_D( u32 fd, u32 fs, u32 ft );
		virtual	void				GenerateSUB_D( u32 fd, u32 fs, u32 ft );
		virtual	void				GenerateMUL_D( u32 fd, u32 fs, u32 ft );
		virtual	void				GenerateDIV_D( u32 fd, u32 fs, u32 ft );
		virtual	void				GenerateSQRT_D( u32 fd, u32 fs );
		virtual	void				GenerateMOV_D( u32 fd, u32 fs );
		virtual	void				GenerateNEG_D( u32 fd, u32 fs );
		virtual	void				GenerateCompare_D( u32 fs, u32 ft, u16 flag );

		virtual	void				GenerateCVT_S_W( u32 fd, u32 fs );
		virtual	void				GenerateCVT_S_D( u32 fd, u32 fs );
		virtual	void				GenerateCVT_D_W( u32 fd, u32 fs );
		virtual	void				GenerateCVT_D_S( u32 fd, u32 fs );
		virtual	void				GenerateCVT_W_S( u32 fd, u32 fs );


	private:
				CJumpLocation		GenerateBranchAlways( CCodeLabel target );
				CJumpLocation		GenerateBranchIfSet( const u32 * p_var, CCodeLabel target );
				CJumpLocation		GenerateBranchIfNotSet( const u32 * p_var, CCodeLabel target );
				CJumpLocation		GenerateBranchIfEqual32( const u32 * p_var, u32 value, CCodeLabel target );
				CJumpLocation		GenerateBranchIfEqual8( const u32 * p_var, u8 value, CCodeLabel target );
				CJumpLocation		GenerateBranchIfNotEqual32( const u32 * p_var, u32 value, CCodeLabel target );
				CJumpLocation		GenerateBranchIfNotEqual8( const u32 * p_var, u8 value, CCodeLabel target );

				EIntelReg			GetMIPSLoRegisterNoLoad(u32 mreg, EIntelReg scratch_to_use);
				EIntelReg			GetMIPSLoRegisterAndLoad(u32 mreg, EIntelReg scratch_to_use);

				enum EUpdateRegOptions
				{
					URO_HI_LEAVE,			// Leave hi bits alone
					URO_HI_SIGN_EXTEND,		// Sign extend from src
					URO_HI_CLEAR,			// Clear hi bits
					URO_HI_SET,				// Set hit bits
				};

				void				UpdateMIPSReg( u32 mreg, EIntelReg src, EUpdateRegOptions options, EIntelReg scratch );

				void				LoadMIPSLo(EIntelReg ireg, u32 mreg);
				void				LoadMIPSHi(EIntelReg ireg, u32 mreg);
				void				StoreMIPSLo(u32 mreg, EIntelReg ireg);
				void				StoreMIPSHi(u32 mreg, EIntelReg ireg);
				void				SetMIPSLo(u32 mreg, u32 data);
				void				SetMIPSHi(u32 mreg, u32 data);

				void				FlushMipsRegs();
				void				RestoreIntelRegs();

				void				GenerateReadWriteAddress(u32 base, u32 offset, u32 xor_val, bool write = false, u32 rt = ~0, bool baseimm = false, bool base_already_loaded = false, EIntelReg destreg = EAX_CODE);
				void				GenerateGenericR4300( u32 address, OpCode op_code, CPU_Instruction p_instruction );

	private:
				bool				mSpCachedInESI;		// Is sp cached in ESI?
				u32					mSetSpPostUpdate;	// Set Sp base counter after this update

				CAssemblyBuffer *	mpPrimary;
				CAssemblyBuffer *	mpSecondary;
};

#endif // CODEGENERATORX86_H_
