/*
Copyright (C) 2001 StrmnNrmn

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
	  
*/

#include "stdafx.h"
#include "TextureCache.h"
#include "ConvertImage.h"
#include "Core\Memory.h"

#include "RDP.h"


//*****************************************************************************
//
//*****************************************************************************
void ConvertRGBA16(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;
	u32 x, y;
	u32 nFiddle;
	
	// Copy of the base pointer
	u8 * pByteSrc = g_pu8RamBase;
	
	if (!pSurf->StartUpdate(&dst))
		return;
	
	if (ti.Swapped)
	{
		u32 dwBaseAddress = ti.Address + (ti.Top * ti.Pitch) + (ti.Left*2);
		u32 * pdwDst = (u32 *)(dst.pSurface);

		for (y = 0; y < ti.Height; y++)
		{
			if ((y&1) == 0)
				nFiddle = 0x2;
			else
				nFiddle = 0x2 | 0x4;
						
			u32 offset = dwBaseAddress;
			
			for (x = 0; x < ti.Width; x++)
			{
				u16 w = *(u16 *)&pByteSrc[offset ^ nFiddle];
				
				pdwDst[x] = Convert555ToRGBA(w);
				
				// Increment word offset to point to the next pixel
				offset += 2;
			}

			dwBaseAddress += ti.Pitch;
			pdwDst = (u32 *)((u8*)pdwDst + dst.Pitch);

		}
	}
	else
	{
		u32 dwBaseAddress = ti.Address + (ti.Top * ti.Pitch) + (ti.Left*2);
		u32 * pdwDst = (u32 *)(dst.pSurface);

		for (y = 0; y < ti.Height; y++)
		{		
			u32 offset = dwBaseAddress;
			
			for (x = 0; x < ti.Width; x++)
			{
				u16 w = *(u16 *)&pByteSrc[offset ^ 0x2];
				
				pdwDst[x] = Convert555ToRGBA(w);
				
				// Increment word offset to point to the next pixel
				offset += 2;
			}

			dwBaseAddress += ti.Pitch;
			pdwDst = (u32 *)((u8*)pdwDst + dst.Pitch);
		}
	}
	
	pSurf->EndUpdate(&dst);
}

//*****************************************************************************
//
//*****************************************************************************
void ConvertRGBA32(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;

	u32 * pSrc = (u32*)(g_pu8RamBase + ti.Address);

	if (!pSurf->StartUpdate(&dst))
		return;
	
	if (ti.Swapped)
	{
		for (u32 y = 0; y < ti.Height; y++)
		{
			if ((y&1) == 0)
			{
				u32 *pD = (u32*)((u8 *)dst.pSurface + y * dst.Pitch);
				u8 *pS = (u8 *)pSrc + (y+ti.Top) * ti.Pitch + (ti.Left*4);
				
				for (u32 x = 0; x < ti.Width; x++)
				{
					*pD++ = MAKE_COLOUR_U32( pS[3], pS[2], pS[1], pS[0] );
					pS+=4;
				}
			}
			else
			{
				
				u32 *pD = (u32 *)((u8 *)dst.pSurface + y * dst.Pitch);
				u8 *pS = (u8 *)pSrc;
				s32 n;
				
				n = (y+ti.Top) * ti.Pitch + (ti.Left*4);
				for (u32 x = 0; x < ti.Width; x++)
				{
					*pD++ = MAKE_COLOUR_U32(pS[(n+3)^0x8],
						pS[(n+2)^0x8],
						pS[(n+1)^0x8],
						pS[(n+0)^0x8]);
					
					n += 4;
				}
			}
		}
	}
	else
	{
		for (u32 y = 0; y < ti.Height; y++)
		{
			u32 *pD = (u32*)((u8 *)dst.pSurface + y * dst.Pitch);
			u8 *pS = (u8 *)pSrc + (y+ti.Top) * ti.Pitch + (ti.Left*4);
			
			for (u32 x = 0; x < ti.Width; x++)
			{
				*pD++ = MAKE_COLOUR_U32( pS[3], pS[2], pS[1], pS[0] );
				pS+=4;
			}
		}
		
	}
	pSurf->EndUpdate(&dst);
	
}

//*****************************************************************************
// E.g. Dear Mario text
// Copy, Score etc
//*****************************************************************************
void ConvertIA4(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;
	u32 nFiddle;

	u8 * pSrc = g_pu8RamBase + ti.Address;

	if (!pSurf->StartUpdate(&dst))
		return;
	
	if (ti.Swapped)
	{
		for (u32 y = 0; y < ti.Height; y++)
		{
			u8 *pD = (u8 *)dst.pSurface + y * dst.Pitch;
			
			// For odd lines, swap words too
			if ((y&1) == 0)
				nFiddle = 0x3;
			else
				nFiddle = 0x7;
			
			
			// This may not work if X is not even?
			u32 dwByteOffset = (y+ti.Top) * ti.Pitch + (ti.Left/2);
			
			// Do two pixels at a time
			for (u32 x = 0; x < ti.Width; x+=2)
			{
				u8 b = pSrc[dwByteOffset ^ nFiddle];
				
				// Even
				*pD++ = TwoToEight[(b & 0xc0) >> 6];
				*pD++ = TwoToEight[(b & 0xc0) >> 6];
				*pD++ = TwoToEight[(b & 0xc0) >> 6];
				*pD++ = TwoToEight[(b & 0x30) >> 4];	
				// Odd
				*pD++ = TwoToEight[(b & 0x0c) >> 2];
				*pD++ = TwoToEight[(b & 0x0c) >> 2];
				*pD++ = TwoToEight[(b & 0x0c) >> 2];
				*pD++ = TwoToEight[(b & 0x03)     ];
				
				dwByteOffset++;	
			}
		}
	}
	else
	{
		for (u32 y = 0; y < ti.Height; y++)
		{
			u8 *pD = (u8 *)dst.pSurface + (y * dst.Pitch);
			
			// This may not work if X is not even?
			u32 dwByteOffset = (y+ti.Top) * ti.Pitch + (ti.Left/2);
			
			// Do two pixels at a time
			for (u32 x = 0; x < ti.Width; x+=2)
			{
				u8 b = pSrc[dwByteOffset ^ 0x3];
				
				// Even
				*pD++ = TwoToEight[(b & 0xc0) >> 6];
				*pD++ = TwoToEight[(b & 0xc0) >> 6];
				*pD++ = TwoToEight[(b & 0xc0) >> 6];
				*pD++ = TwoToEight[(b & 0x30) >> 4];	
				// Odd
				*pD++ = TwoToEight[(b & 0x0c) >> 2];
				*pD++ = TwoToEight[(b & 0x0c) >> 2];
				*pD++ = TwoToEight[(b & 0x0c) >> 2];
				*pD++ = TwoToEight[(b & 0x03)     ];
				
				dwByteOffset++;
				
			}
		}	
	}
	
	pSurf->EndUpdate(&dst);
	
}

//*****************************************************************************
// E.g Mario's head textures
//*****************************************************************************
void ConvertIA8(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;
	u32 nFiddle;

	//u8 * pSrc = g_pu8RamBase + ti.Address;

	if (!pSurf->StartUpdate(&dst))
		return;
	
	if (ti.Swapped)
	{
		for (u32 y = 0; y < ti.Height; y++)
		{
			// For odd lines, swap words too
			if ((y&1) == 0)
				nFiddle = 0x3;
			else
				nFiddle = 0x7;
			
			u8 *pD = (u8 *)dst.pSurface + y * dst.Pitch;
			// Points to current byte
			u32 dwByteOffset = ((y+ti.Top) * ti.Pitch) + ti.Left;
			
			for (u32 x = 0; x < ti.Width; x++)
			{
				//u8 b = pSrc[dwByteOffset ^ nFiddle];
				u8 b = g_pu8RamBase[(dwByteOffset + ti.Address) ^ nFiddle];
				
				
				*pD++ = FourToEight[(b & 0xf0)>>4];
				*pD++ = FourToEight[(b & 0xf0)>>4];
				*pD++ = FourToEight[(b & 0xf0)>>4];
				*pD++ = FourToEight[(b & 0x0f)   ];
				
				dwByteOffset++;
			}
		}		
	}
	else
	{
		for (u32 y = 0; y < ti.Height; y++)
		{
			u8 *pD = (u8 *)dst.pSurface + y * dst.Pitch;
			
			// Points to current byte
			u32 dwByteOffset = ((y+ti.Top) * ti.Pitch) + ti.Left;
			
			for (u32 x = 0; x < ti.Width; x++)
			{
				//u8 b = pSrc[dwByteOffset ^ 0x3];
				u8 b = g_pu8RamBase[(dwByteOffset + ti.Address) ^ 0x3];
				
				*pD++ = FourToEight[(b & 0xf0)>>4];
				*pD++ = FourToEight[(b & 0xf0)>>4];
				*pD++ = FourToEight[(b & 0xf0)>>4];
				*pD++ = FourToEight[(b & 0x0f)   ];
				
				dwByteOffset++;
			}
		}
	}	
	
	pSurf->EndUpdate(&dst);
	
}

//*****************************************************************************
// E.g. camera's clouds, shadows
//*****************************************************************************
void ConvertIA16(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;
	u32 nFiddle;
	
	u16 * pSrc = (u16*)(g_pu8RamBase + ti.Address);
	u8 * pByteSrc = (u8 *)pSrc;
	
	if (!pSurf->StartUpdate(&dst))
		return;
	
	if (ti.Swapped)
	{
		
		for (u32 y = 0; y < ti.Height; y++)
		{
			u8 *pD = (u8 *)dst.pSurface + y * dst.Pitch;
			
			if ((y&1) == 0)
				nFiddle = 0x2;
			else
				nFiddle = 0x4 | 0x2;
			
			// Points to current word
			u32 dwWordOffset = ((y+ti.Top) * ti.Pitch) + (ti.Left * 2);
			
			for (u32 x = 0; x < ti.Width; x++)
			{
				u16 w = *(u16 *)&pByteSrc[dwWordOffset^nFiddle];
				
				*pD++ = (u8)(w >> 8);
				*pD++ = (u8)(w >> 8);
				*pD++ = (u8)(w >> 8);
				*pD++ = (u8)(w & 0xFF);
				
				dwWordOffset += 2;
			}
		}		
	}
	else
	{
		for (u32 y = 0; y < ti.Height; y++)
		{
			u8 *pD = (u8 *)dst.pSurface + y * dst.Pitch;
			
			// Points to current word
			u32 dwWordOffset = ((y+ti.Top) * ti.Pitch) + (ti.Left * 2);
			
			for (u32 x = 0; x < ti.Width; x++)
			{
				u16 w = *(u16 *)&pByteSrc[dwWordOffset^0x2];
				
				*pD++ = (u8)(w >> 8);
				*pD++ = (u8)(w >> 8);
				*pD++ = (u8)(w >> 8);
				*pD++ = (u8)(w & 0xFF);
				
				dwWordOffset += 2;
			}
		}		
		
	}
	
	
	pSurf->EndUpdate(&dst);
}



//*****************************************************************************
// Used by MarioKart
//*****************************************************************************
void ConvertI4(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;
	u32 nFiddle;
	
	u8 * pSrc = g_pu8RamBase + ti.Address;

	if (!pSurf->StartUpdate(&dst))
		return;
	
	if (ti.Swapped)
	{
		
		for (u32 y = 0; y < ti.Height; y++)
		{
			u8 *pD = (u8 *)dst.pSurface + y * dst.Pitch;
			
			// Might not work with non-even starting X
			u32 dwByteOffset = ((y+ti.Top) * ti.Pitch) + (ti.Left / 2);
			
			// For odd lines, swap words too
			if ((y&1) == 0)
				nFiddle = 0x3;
			else
				nFiddle = 0x7;
			
			for (u32 x = 0; x < ti.Width; x+=2)
			{
				u8 b = pSrc[dwByteOffset ^ nFiddle];
				
				// Even
				*pD++ = FourToEight[(b & 0xF0)>>4];	// Other implementations seem to or in (b&0xF0)>>4
				*pD++ = FourToEight[(b & 0xF0)>>4]; // why?
				*pD++ = FourToEight[(b & 0xF0)>>4];
				*pD++ = FourToEight[(b & 0xF0)>>4];	
				// Odd
				*pD++ = FourToEight[(b & 0x0F)];
				*pD++ = FourToEight[(b & 0x0F)];
				*pD++ = FourToEight[(b & 0x0F)];
				*pD++ = FourToEight[(b & 0x0F)];
				
				dwByteOffset++;
			}
			
		}	
		
	}
	else
	{
		
		for (u32 y = 0; y < ti.Height; y++)
		{
			u8 *pD = (u8 *)dst.pSurface + y * dst.Pitch;
			
			// Might not work with non-even starting X
			u32 dwByteOffset = ((y+ti.Top) * ti.Pitch) + (ti.Left / 2);
			
			for (u32 x = 0; x < ti.Width; x+=2)
			{
				u8 b = pSrc[dwByteOffset ^ 0x3];
				
				// Even
				*pD++ = FourToEight[(b & 0xF0)>>4];	// Other implementations seem to or in (b&0xF0)>>4
				*pD++ = FourToEight[(b & 0xF0)>>4]; // why?
				*pD++ = FourToEight[(b & 0xF0)>>4];
				*pD++ = FourToEight[(b & 0xF0)>>4];	
				// Odd
				*pD++ = FourToEight[(b & 0x0F)];
				*pD++ = FourToEight[(b & 0x0F)];
				*pD++ = FourToEight[(b & 0x0F)];
				*pD++ = FourToEight[(b & 0x0F)];
				
				dwByteOffset++;
			}
		}
	}
	pSurf->EndUpdate(&dst);
}

//*****************************************************************************
// Used by MarioKart
//*****************************************************************************
void ConvertI8(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;
	u32 nFiddle;
	
	if (!pSurf->StartUpdate(&dst))
		return;
	
	if (ti.Swapped)
	{
		for (u32 y = 0; y < ti.Height; y++)
		{
			if ((y&1) == 0)
				nFiddle = 0x3;
			else
				nFiddle = 0x7;
			
			u8 *pD = (u8 *)dst.pSurface + y * dst.Pitch;
			
			u32 dwByteOffset = ((y+ti.Top) * ti.Pitch) + ti.Left;
			
			for (u32 x = 0; x < ti.Width; x++)
			{
				//u8 b = pSrc[dwByteOffset ^ nFiddle];
				u8 b = g_pu8RamBase[(dwByteOffset + ti.Address) ^ nFiddle];
				
				*pD++ = b;
				*pD++ = b;
				*pD++ = b;
				*pD++ = b;		// Alpha not 255?
				
				dwByteOffset++;
			}
		}	
	}
	else
	{
		for (u32 y = 0; y < ti.Height; y++)
		{
			u8 *pD = (u8 *)dst.pSurface + y * dst.Pitch;
			
			u32 dwByteOffset = ((y+ti.Top) * ti.Pitch) + ti.Left;
			
			for (u32 x = 0; x < ti.Width; x++)
			{
				//u8 b = pSrc[dwByteOffset ^ 0x3];
				u8 b = g_pu8RamBase[(dwByteOffset + ti.Address) ^ 0x3];
				
				*pD++ = b;
				*pD++ = b;
				*pD++ = b;
				*pD++ = b;		// Alpha not 255?
				
				dwByteOffset++;
			}
		}	
		
	}
	pSurf->EndUpdate(&dst);
	
}


//*****************************************************************************
// Used by Starfox intro
//*****************************************************************************
void ConvertCI4_RGBA16(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;
	u32 nFiddle;

	u8 * pSrc = g_pu8RamBase + ti.Address;
	u16 * pPal = (u16 *)&gTextureMemory[ ti.TmemPalAddress << 3 ];

	if (!pSurf->StartUpdate(&dst))
		return;
	
	if (ti.Swapped)
	{
		
		for (u32 y = 0; y < ti.Height; y++)
		{
			if ((y&1) == 0)
				nFiddle = 0x3;
			else
				nFiddle = 0x7;
			
			
			u32 * pD = (u32 *)((u8 *)dst.pSurface + y * dst.Pitch);
			
			u32 dwByteOffset = ((y+ti.Top) * ti.Pitch) + (ti.Left / 2);
			
			for (u32 x = 0; x < ti.Width; x+=2)
			{
				u8 b = pSrc[dwByteOffset ^ nFiddle];
				
				u8 bhi = (b&0xf0)>>4;
				u8 blo = (b&0x0f);
				
				pD[0] = Convert555ToRGBA(pPal[bhi^0x1]);	// Remember palette is in different endian order!
				pD[1] = Convert555ToRGBA(pPal[blo^0x1]);	// Remember palette is in different endian order!
				pD+=2;
				
				dwByteOffset++;
			}
		}	
		
	}
	else
	{
		
		for (u32 y = 0; y < ti.Height; y++)
		{
			u32 * pD = (u32 *)((u8 *)dst.pSurface + y * dst.Pitch);
			
			u32 dwByteOffset = ((y+ti.Top) * ti.Pitch) + (ti.Left / 2);
			
			for (u32 x = 0; x < ti.Width; x+=2)
			{
				u8 b = pSrc[dwByteOffset ^ 0x3];
				
				u8 bhi = (b&0xf0)>>4;
				u8 blo = (b&0x0f);
				
				pD[0] = Convert555ToRGBA(pPal[bhi^0x1]);	// Remember palette is in different endian order!
				pD[1] = Convert555ToRGBA(pPal[blo^0x1]);	// Remember palette is in different endian order!
				pD+=2;
				
				dwByteOffset++;
			}
		}	
		
	}
	pSurf->EndUpdate(&dst);
}

//*****************************************************************************
// Used by Starfox intro
//*****************************************************************************
void ConvertCI4_IA16(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;
	u32 nFiddle;
	
	u8 * pSrc = g_pu8RamBase + ti.Address;
	u16 * pPal = (u16 *)&gTextureMemory[ ti.TmemPalAddress << 3 ];

	if (!pSurf->StartUpdate(&dst))
		return;
	
	if (ti.Swapped)
	{
		
		for (u32 y = 0; y < ti.Height; y++)
		{
			if ((y&1) == 0)
				nFiddle = 0x3;
			else
				nFiddle = 0x7;
			
			
			u32 * pD = (u32 *)((u8 *)dst.pSurface + y * dst.Pitch);
			
			u32 dwByteOffset = ((y+ti.Top) * ti.Pitch) + (ti.Left / 2);
			
			for (u32 x = 0; x < ti.Width; x+=2)
			{
				u8 b = pSrc[dwByteOffset ^ nFiddle];
				
				u8 bhi = (b&0xf0)>>4;
				u8 blo = (b&0x0f);
				
				pD[0] = ConvertIA16ToRGBA(pPal[bhi^0x1]);	// Remember palette is in different endian order!
				pD[1] = ConvertIA16ToRGBA(pPal[blo^0x1]);	// Remember palette is in different endian order!
				pD+=2;
				
				dwByteOffset++;
			}
		}	
		
	}
	else
	{
		
		for (u32 y = 0; y < ti.Height; y++)
		{
			u32 * pD = (u32 *)((u8 *)dst.pSurface + y * dst.Pitch);
			
			u32 dwByteOffset = ((y+ti.Top) * ti.Pitch) + (ti.Left / 2);
			
			for (u32 x = 0; x < ti.Width; x+=2)
			{
				u8 b = pSrc[dwByteOffset ^ 0x3];
				
				u8 bhi = (b&0xf0)>>4;
				u8 blo = (b&0x0f);
				
				pD[0] = ConvertIA16ToRGBA(pPal[bhi^0x1]);	// Remember palette is in different endian order!
				pD[1] = ConvertIA16ToRGBA(pPal[blo^0x1]);	// Remember palette is in different endian order!
				pD+=2;
				
				dwByteOffset++;
			}
		}	
		
	}
	pSurf->EndUpdate(&dst);
}




//*****************************************************************************
// Used by MarioKart for Cars etc
//*****************************************************************************
void ConvertCI8_RGBA16(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;
	u32 nFiddle;
	
	u8 * pSrc = g_pu8RamBase + ti.Address;
	u16 * pPal = (u16 *)&gTextureMemory[ ti.TmemPalAddress << 3 ];

	if (!pSurf->StartUpdate(&dst))
		return;
	
	if (ti.Swapped)
	{	
		for (u32 y = 0; y < ti.Height; y++)
		{
			if ((y&1) == 0)
				nFiddle = 0x3;
			else
				nFiddle = 0x7;
			
			u32 *pD = (u32 *)((u8 *)dst.pSurface + y * dst.Pitch);
			
			u32 dwByteOffset = ((y+ti.Top) * ti.Pitch) + ti.Left;
			
			for (u32 x = 0; x < ti.Width; x++)
			{
				u8 b = pSrc[dwByteOffset ^ nFiddle];
				
				*pD++ = Convert555ToRGBA(pPal[b^0x1]);	// Remember palette is in different endian order!
				
				dwByteOffset++;
			}
		}	
		
		
	}
	else
	{
		
		for (u32 y = 0; y < ti.Height; y++)
		{
			u32 *pD = (u32 *)((u8 *)dst.pSurface + y * dst.Pitch);
			
			u32 dwByteOffset = ((y+ti.Top) * ti.Pitch) + ti.Left;
			
			for (u32 x = 0; x < ti.Width; x++)
			{
				u8 b = pSrc[dwByteOffset ^ 0x3];
				
				*pD++ = Convert555ToRGBA(pPal[b^0x1]);	// Remember palette is in different endian order!
				
				dwByteOffset++;
			}
		}
	}
	pSurf->EndUpdate(&dst);
	
}



//*****************************************************************************
// Used by MarioKart for Cars etc
//*****************************************************************************
void ConvertCI8_IA16(CTexture *pSurf, const TextureInfo & ti)
{
	DrawInfo dst;
	u32 nFiddle;
	
	u8 * pSrc = g_pu8RamBase + ti.Address;
	u16 * pPal = (u16 *)&gTextureMemory[ ti.TmemPalAddress << 3 ];

	if (!pSurf->StartUpdate(&dst))
		return;
	
	if (ti.Swapped)
	{
		for (u32 y = 0; y < ti.Height; y++)
		{
			if ((y&1) == 0)
				nFiddle = 0x3;
			else
				nFiddle = 0x7;
			
			u32 *pD = (u32 *)((u8 *)dst.pSurface + y * dst.Pitch);
			
			u32 dwByteOffset = ((y+ti.Top) * ti.Pitch) + ti.Left;
			
			for (u32 x = 0; x < ti.Width; x++)
			{
				u8 b = pSrc[dwByteOffset ^ nFiddle];
				
				*pD++ = ConvertIA16ToRGBA(pPal[b^0x1]);	// Remember palette is in different endian order!
				
				dwByteOffset++;
			}
		}	
	}
	else
	{
		for (u32 y = 0; y < ti.Height; y++)
		{
			u32 *pD = (u32 *)((u8 *)dst.pSurface + y * dst.Pitch);
			
			u32 dwByteOffset = ((y+ti.Top) * ti.Pitch) + ti.Left;
			
			for (u32 x = 0; x < ti.Width; x++)
			{
				u8 b = pSrc[dwByteOffset ^ 0x3];
				
				*pD++ = ConvertIA16ToRGBA(pPal[b^0x1]);	// Remember palette is in different endian order!
				
				dwByteOffset++;
			}
		}
	}
	pSurf->EndUpdate(&dst);
}


