//
//	Taken from Rice's adaptation of the Daedalus graphics plugin
//

/*
Copyright (C) 2005 StrmnNrmn
Copyright (C) 2003 Rice1964

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "PolyClipper.h"
#include "PSPColour.h"
#include "FFXMaths.h"

#include "Vector3.h"

#include <vector>

//#define DEBUG_COLOUR_CLIPPED_POLYS
//#define DEBUG_POLYCLIPPER

static u32 gScreenWidth = 480;
static u32 gScreenHeight = 272;

//////////////////////////////////////////////////////////////////////////
/*
 *	Manual vertex clipper
 */
//////////////////////////////////////////////////////////////////////////
typedef f32 floatx;			// f64 is VERY expensive on PSP

SClipVertex		g_clippedVtxBuffer[ 1000 ];
u32				g_clippedVtxCount = 0;
static bool		gFogEnabled( false );				// XXXXX!
static bool		gZCompare( false );

const u32		gAllNegWColour( MAKE_COLOUR_U32( 255,0,255 ) );
const u32		gAllNegZColour( MAKE_COLOUR_U32( 255,0,0 ) );
const u32		gAllPosZColour( MAKE_COLOUR_U32( 0,255,0 ) );

bool _finite( floatx )
{
	return true;
}


floatx	InterpSegment(const v3 & a, const v3 & b, const v3 & c)
{
	float	denom_x( c.x - b.x );

	if(fabsf(denom_x) > 0.0f)
	{
		return (a.x - b.x) / denom_x;
	}

	float	denom_y( c.y - b.y );

	if(fabsf(denom_y) > 0.0f)
	{
		return (a.y - b.y) / denom_y;
	}

	// XXXX
	return 0.0f;
}



void CopyVertexData(s32 oriidx, const SClipVertex *oribuf, s32 dstidx, SClipVertex *dstbuf)
{
	memcpy(dstbuf+dstidx,oribuf+oriidx, sizeof(SClipVertex));
}

void CopyTriangleDataWithColour(s32 oriidx, const SClipVertex *oribuf, s32 dstidx, SClipVertex *dstbuf, u32 colour )
{
	for(u32 i = 0; i < 3; ++i)
	{
		memcpy(dstbuf+dstidx+i,oribuf+oriidx+i, sizeof(SClipVertex));

		dstbuf[dstidx+i].r = COL_GETR( colour );
		dstbuf[dstidx+i].g = COL_GETG( colour );
		dstbuf[dstidx+i].b = COL_GETB( colour );
		dstbuf[dstidx+i].a = 255;
	}
}


#define interp(a,b,r)	(((a)-(r)*(b))/	(1-(r)))

void SetVtx(SClipVertex &v, SClipVertex &vtx1, SClipVertex &vtx2, f32 r)
{
	v.x = interp(vtx1.x,vtx2.x,r);
	v.y = interp(vtx1.y,vtx2.y,r);
	//v.z = (vtx1.z-r*vtx2.z)/(1-r);
	v.specular = vtx2.specular; //fix me here
	v.r = (u8)(interp((s32)vtx1.r,(s32)vtx2.r,r));
	v.g = (u8)(interp((s32)vtx1.g,(s32)vtx2.g,r));
	v.b = (u8)(interp((s32)vtx1.b,(s32)vtx2.b,r));
	v.a = (u8)(interp((s32)vtx1.a,(s32)vtx2.a,r));

	for( u32 i=0; i<SClipVertex::NUM_TEX_COORDS; i++ )
	{
		v.tcoord[i].u = interp(vtx1.tcoord[i].u,vtx2.tcoord[i].u,r);
		v.tcoord[i].v = interp(vtx1.tcoord[i].v,vtx2.tcoord[i].v,r);
	}
}



void SwapVertexPos(SClipVertex * p_vtx_buffer, s32 firstidx)
{
	SClipVertex &vtx1 = p_vtx_buffer[firstidx];
	SClipVertex &vtx2 = p_vtx_buffer[firstidx+1];
	SClipVertex &vtx3 = p_vtx_buffer[firstidx+2];

	if( vtx1.rhw >= vtx2.rhw && vtx1.rhw >= vtx3.rhw ) return;

	SClipVertex tempv;
	memcpy(&tempv,&p_vtx_buffer[firstidx], sizeof(SClipVertex));

	if( vtx2.rhw > vtx1.rhw && vtx2.rhw >= vtx3.rhw )
	{
		// vtx2 is the largest one
		memcpy(&p_vtx_buffer[firstidx],&p_vtx_buffer[firstidx+1], sizeof(SClipVertex));
		memcpy(&p_vtx_buffer[firstidx+1],&p_vtx_buffer[firstidx+2], sizeof(SClipVertex));
		memcpy(&p_vtx_buffer[firstidx+2],&tempv, sizeof(SClipVertex));
	}
	else
	{
		memcpy(&p_vtx_buffer[firstidx],&p_vtx_buffer[firstidx+2], sizeof(SClipVertex));
		memcpy(&p_vtx_buffer[firstidx+2],&p_vtx_buffer[firstidx+1], sizeof(SClipVertex));
		memcpy(&p_vtx_buffer[firstidx+1],&tempv, sizeof(SClipVertex));
	}
}

/************************************************************************/
/* Manually clipping vertexes                                           */
/* DirectX won't clip transformed vertex, unless the vertex is transfor-*/
/* med by DirectX, so we have to do vertex clipping ourself. Otherwise  */
/* this plugin works very bad on newest video cards, like ATI Radeons.  */
/************************************************************************/
bool Clip1TriangleForNegW(const SClipVertex &vtx1, const SClipVertex &vtx2, const SClipVertex &vtx3, s32 &dstidx);
bool Clip1TriangleForZ(const SClipVertex &vtx1, const SClipVertex &vtx2, const SClipVertex &vtx3, s32 &dstidx);

void ClipVertexes( /*const */SClipVertex * p_vtx_buffer, u32 num_vertices, bool z_cmp, bool fog_enabled )
{
	s32 dstidx = 0;

/*
	for( u32 i=0; i<num_vertices/3; i++)
	{
		s32 firstidx = i*3;
		const SClipVertex &vtx1 = p_vtx_buffer[firstidx];
		const SClipVertex &vtx2 = p_vtx_buffer[firstidx+1];
		const SClipVertex &vtx3 = p_vtx_buffer[firstidx+2];

		CopyVertexData(firstidx,   p_vtx_buffer, dstidx++, g_clippedVtxBuffer);
		CopyVertexData(firstidx+1, p_vtx_buffer, dstidx++, g_clippedVtxBuffer);
		CopyVertexData(firstidx+2, p_vtx_buffer, dstidx++, g_clippedVtxBuffer);
	}

	g_clippedVtxCount = dstidx;
	return;
*/
	gZCompare = z_cmp;
	gFogEnabled = fog_enabled;

	// transverse the list of transformed vertex for each triangles
	// - If z values of all 3 vertexes of the triangle are greater than 1, then ignore this triangle
	// - If non z value of all 3 vertexes of the triangle is greater than 1, then this triangle is ok
	// - If one or two z values are greater than 1, then split this triangle to 2 triangles


	for( u32 i=0; i<num_vertices/3; i++)
	{
		s32 firstidx = i*3;
		const SClipVertex &vtx1 = p_vtx_buffer[firstidx];
		const SClipVertex &vtx2 = p_vtx_buffer[firstidx+1];
		const SClipVertex &vtx3 = p_vtx_buffer[firstidx+2];

		if( vtx1.rhw < 0 && vtx2.rhw < 0 && vtx3.rhw < 0 )
		{
			// All negative rhws - Skip this triangle
			#ifdef DEBUG_COLOUR_CLIPPED_POLYS
			CopyTriangleDataWithColour( firstidx, p_vtx_buffer, dstidx, g_clippedVtxBuffer, gAllNegWColour );	dstidx += 3;
			#endif
			continue;
		}

		if( vtx1.rhw >= 0 && vtx2.rhw >= 0 && vtx3.rhw >= 0 )	
		{
			//
			//	All positive rhws
			//
			if( vtx1.z < 0 || vtx2.z < 0 || vtx3.z < 0 )
			{
				if( !gZCompare )
				{
					// Do nothing about this triangle
					CopyVertexData(firstidx,   p_vtx_buffer, dstidx++, g_clippedVtxBuffer);
					CopyVertexData(firstidx+1, p_vtx_buffer, dstidx++, g_clippedVtxBuffer);
					CopyVertexData(firstidx+2, p_vtx_buffer, dstidx++, g_clippedVtxBuffer);
					continue;
				}
				else if( vtx1.z < 0 && vtx2.z < 0 && vtx3.z < 0 )
				{
					#ifdef DEBUG_COLOUR_CLIPPED_POLYS
					CopyTriangleDataWithColour( firstidx, p_vtx_buffer, dstidx, g_clippedVtxBuffer, gAllNegZColour );	dstidx += 3;
					#endif

					continue;
				}
				else
				{
					// Clipping this triangle for z
					#ifdef DEBUG_POLYCLIPPER
						if( pauseAtNext && logTriangles ) 
						{
							TRACE1("Clip triangle %d against near plane", i);
							TRACE3("Old triangle z values: %f, %f, %f", vtx1.z, vtx2.z, vtx3.z);
						}
					#endif
					Clip1TriangleForZ(vtx1, vtx2, vtx3, dstidx);
					continue;
				}
			}
			else if( vtx1.z > 1 || vtx2.z > 1 || vtx3.z > 1 )
			{
				if( vtx1.z > 1 && vtx2.z > 1 && vtx3.z > 1 )
				{
					#ifdef DEBUG_COLOUR_CLIPPED_POLYS
					CopyTriangleDataWithColour( firstidx, p_vtx_buffer, dstidx, g_clippedVtxBuffer, gAllPosZColour );	dstidx += 3;
					#endif

					continue;
				}
				else
				{
					// Clipping this triangle for z
					#ifdef DEBUG_POLYCLIPPER
						if( pauseAtNext && logTriangles ) 
						{
							TRACE1("Clip triangle %d against far plane", i);
							TRACE3("Old triangle z values: %f, %f, %f", vtx1.z, vtx2.z, vtx3.z);
						}
					#endif
					Clip1TriangleForZ(vtx1, vtx2, vtx3, dstidx);
					continue;
				}
			}
			else
			{
				// Do nothing about this triangle
				CopyVertexData(firstidx,   p_vtx_buffer, dstidx++, g_clippedVtxBuffer);
				CopyVertexData(firstidx+1, p_vtx_buffer, dstidx++, g_clippedVtxBuffer);
				CopyVertexData(firstidx+2, p_vtx_buffer, dstidx++, g_clippedVtxBuffer);
			}
			continue;
		}
		else
		{
			#ifdef DEBUG_POLYCLIPPER
				if( pauseAtNext && logTriangles ) 
				{
					TRACE1("Clip triangle %d for negative w", i);
					TRACE3("Old triangle w values: %f, %f, %f", vtx1.rhw, vtx2.rhw, vtx3.rhw);
				}
			#endif
			SwapVertexPos(p_vtx_buffer, firstidx);

			#ifdef DEBUG_POLYCLIPPER
				if( pauseAtNext && logTriangles ) 
				{
					TRACE3("Old triangle w values: %f, %f, %f", vtx1.rhw, vtx2.rhw, vtx3.rhw);
				}
			#endif

			Clip1TriangleForNegW(vtx1, vtx2, vtx3, dstidx);
		}
	}

	g_clippedVtxCount = dstidx;
}

struct LineEquationType
{
	floatx d;
	floatx x;
	floatx y;
};
LineEquationType lines[3];	// In the format: Nx, Ny, d

floatx EvaLine(const LineEquationType &li, floatx x, floatx y)
{
	return li.x*x + li.y*y + li.d;
}


inline v3 Split( const v3 &a, const v3 &b, const LineEquationType &line  )
{
	floatx aDot = (a.x*line.x + a.y*line.y);
	floatx bDot = (b.x*line.x + b.y*line.y);

	floatx scale = ( - line.d - aDot) / ( bDot - aDot );

	return a + ((b - a) * (f32)scale );
}

// Clipping using the Sutherland-Hodgeman algorithm
bool ClipFor1LineXY( const std::vector<v3> &in, std::vector<v3> &out, const LineEquationType &line )
{
	s32 insize = in.size();
	s32 thisInd=insize-1;
	s32 nextInd=0;

	floatx thisRes = EvaLine( line, in[thisInd].x, in[thisInd].y );
	floatx nextRes;

	out.clear();

	for( nextInd=0; nextInd<insize; nextInd++ )
	{
		nextRes = EvaLine( line, in[nextInd].x, in[nextInd].y );

		if( thisRes >= 0 )
		{
			// Add the point
			out.push_back(in[thisInd]);
		}

		if( ( thisRes < 0 && nextRes >= 0 ) || ( thisRes >= 0 && nextRes < 0 ) )
		{
			// Add the split point
			out.push_back( Split(in[thisInd], in[nextInd], line ));
		}

		thisInd = nextInd;
		thisRes = nextRes;
	}
	if( (s32)out.size() >= insize )
	{
		return true;
	}
	return false;
}
// Clipping using the Sutherland-Hodgeman algorithm
bool ClipFor1LineZ( const std::vector<v3> &in, std::vector<v3> &out, bool nearplane )
{
	const f32 nearz = 1e-4f;
	const f32 farz = 1-1e-4f;

	s32 insize = in.size();
	s32 thisInd=insize-1;
	s32 nextInd=0;

	bool thisRes = nearplane ? (in[thisInd].z >= nearz) : (in[thisInd].z <= farz) ;
	bool nextRes;

	out.clear();

	for( nextInd=0; nextInd<insize; nextInd++ )
	{
		nextRes = nearplane ? (in[nextInd].z >= nearz) : (in[nextInd].z <= farz) ;

		if( thisRes )
		{
			// Add the point
			out.push_back(in[thisInd]);
		}

		if( ( !thisRes && nextRes ) || ( thisRes && !nextRes ) )
		{
			// Add the split point
			v3 newvtx;
			const v3 &vtx1 = in[thisInd];
			const v3 &vtx2 = in[nextInd];

			newvtx.z = nearplane ? nearz : farz;

			f32 r;
			f32	z_diff( vtx1.z - vtx2.z );
			if(fabsf(z_diff) > 0.0f)
			{
				r = (vtx1.z - newvtx.z) / z_diff;
			}
			else
			{
				// z is the same!
				r = 0.5f;
			}
			newvtx.x = vtx1.x - r*(vtx1.x-vtx2.x);
			newvtx.y = vtx1.y - r*(vtx1.y-vtx2.y);

			out.push_back( newvtx );
		}

		thisInd = nextInd;
		thisRes = nextRes;
	}
	if( (s32)out.size() >= insize )
	{
		return true;
	}
	return false;
}

void Create1LineEq(LineEquationType &l, const SClipVertex &vtx1, const SClipVertex &vtx2, const SClipVertex &vtx3)
{
	// Line between (x1,y1) to (x2,y2)
	l.x = vtx2.y-vtx1.y;
	l.y = vtx1.x-vtx2.x;
	l.d = -( l.x*vtx2.x + l.y*vtx2.y );
	if( EvaLine(l,vtx3.x,vtx3.y)*vtx3.rhw<0 )
	{
		l.x = -l.x;
		l.y = -l.y;
		l.d = -l.d;
	}
}

f32 inline interp3p(f32 a, f32 b, f32 c, floatx r1, floatx r2)
{
	floatx		bc( b + (c-b)*r2 );

	return f32( a + (bc-a)*r1 );
}

void Interp1PtZ(v3 &v, const SClipVertex &vtx1, const SClipVertex &vtx2, const SClipVertex &vtx3)
{
	if( fabsf(v.x-vtx1.x)<1 && fabsf(v.y-vtx1.y) < 1 )
	{
		v.z = vtx1.z;
		return;
	}
	if( fabsf(v.x-vtx2.x)<1 && fabsf(v.y-vtx2.y) < 1 )
	{
		v.z = vtx2.z;
		return;
	}
	if( fabsf(v.x-vtx3.x)<1 && fabsf(v.y-vtx3.y) < 1 )
	{
		v.z = vtx3.z;
		return;
	}

	LineEquationType line;
	Create1LineEq(line, vtx2, vtx3, vtx1);

	v3 tempv1(vtx1.x, vtx1.y, 0);
	floatx aDot = (v.x*line.x + v.y*line.y);
	floatx bDot = (vtx1.x*line.x + vtx1.y*line.y);

	floatx scale1 = ( - line.d - aDot) / ( bDot - aDot );

	v3 tempv( v + ((f32)scale1 * (tempv1 - v)) );

	floatx	s1( InterpSegment( v, v3(vtx1.x, vtx1.y, vtx1.z), tempv ) );
	floatx	s2( InterpSegment( tempv, v3(vtx2.x, vtx2.y, vtx2.z), v3(vtx3.x, vtx3.y, vtx3.z) ) );

	v.z = interp3p(vtx1.z,vtx2.z,vtx3.z,s1,s2);

#ifdef DEBUG_POLYCLIPPER
	if( !_finite(v.z) )
	{
		TRACE0("Z value is #INF, check me");
	}

	if( fabsf(v.z) > 10 )
	{
		TRACE0("Large Z");
	}
#endif
}

bool Interp1Pt(v3 &v, const SClipVertex &vtx1, const SClipVertex &vtx2, const SClipVertex &vtx3, SClipVertex &out)
{
	LineEquationType line;
	Create1LineEq(line, vtx2, vtx3, vtx1);

	v3 tempv1(vtx1.x, vtx1.y, 0);

	floatx aDot = (v.x*line.x + v.y*line.y);
	floatx bDot = (vtx1.x*line.x + vtx1.y*line.y);

	floatx scale1 = ( - line.d - aDot) / ( bDot - aDot );

	v3 tempv;
	tempv = v + ((f32)scale1 * (tempv1 - v));

	floatx	s1( InterpSegment( v, v3(vtx1.x, vtx1.y, vtx1.z), tempv ) );
	floatx	s2( InterpSegment( tempv, v3(vtx2.x, vtx2.y, vtx2.z), v3(vtx3.x, vtx3.y, vtx3.z) ) );

	if( !_finite(s1) || !_finite(s2) )
	{
		memcpy(&out, &vtx3, sizeof(SClipVertex) );

#ifdef DEBUG_POLYCLIPPER
		TRACE0("s1 or s2 is still infinite, it is a coplaner triangle");
#endif
		return false;
	}

	//out.x = interp3p(vtx1.x,vtx2.x,vtx3.x,s1,s2);
	out.x = v.x;
	out.y = v.y;
	//out.y = interp3p(vtx1.y,vtx2.y,vtx3.y,s1,s2);
	out.z = interp3p(vtx1.z,vtx2.z,vtx3.z,s1,s2);
	out.rhw = interp3p(vtx1.rhw,vtx2.rhw,vtx3.rhw,s1,s2);
#ifdef DEBUG_POLYCLIPPER
	if( !_finite(out.z) || !_finite(out.rhw) )
	{
		TRACE0("Z or RHW value is #INF, check me");
	}
#endif


#ifdef DEBUG_POLYCLIPPER
	if( out.rhw < 0 )	
	{
		DebuggerAppendMsg("Warning: rhw<0");
		DebuggerAppendMsg("nx=[%f %f %f %f];\n",vtx1.x, vtx2.x, vtx3.x, vtx1.x);
		DebuggerAppendMsg("ny=[%f %f %f %f];\n",vtx1.y, vtx2.y, vtx3.y, vtx1.y);
		DebuggerAppendMsg("nz=[%f %f %f %f];\n",vtx1.z, vtx2.z, vtx3.z, vtx1.z);
		DebuggerAppendMsg("newx=[%f %f %f];\n",vtx1.x, v.x, tempv.x);
		DebuggerAppendMsg("newy=[%f %f %f];\n",vtx1.y, v.y, tempv.y);
		DebuggerAppendMsg("newz=[%f %f %f];\n",vtx1.z, v.z, tempv.z);
	}
	if( out.z < 0 )	
		DebuggerAppendMsg("Warning: z<0");
	if( out.z > 1 )	
		DebuggerAppendMsg("Warning: z>1");
#endif



	out.specular = vtx2.specular; //fix me here
	if( gFogEnabled )
	{
		f32 f1 = (vtx1.specular>>24)*vtx1.rhw;
		f32 f2 = (vtx2.specular>>24)*vtx2.rhw;
		f32 f3 = (vtx3.specular>>24)*vtx3.rhw;
		f32 f = interp3p(f1,f2,f3,s1,s2)/out.rhw;	
		if( f < 0 )	f = 0;
		else if( f > 255 ) f = 255;
		u32 fb = (u8)f;
		out.specular &= 0x00FFFFFF;
		out.specular |= (fb<<24);
	}

	f32 r = interp3p(vtx1.r*vtx1.rhw,vtx2.r*vtx2.rhw,vtx3.r*vtx3.rhw,s1,s2)/out.rhw;	
	if( r<0 )	r=0; 
	if( r>255 )	r=255;
	out.r = (u8)r;
	f32 g = interp3p(vtx1.g*vtx1.rhw,vtx2.g*vtx2.rhw,vtx3.g*vtx3.rhw,s1,s2)/out.rhw;
	if( g<0 )	g=0; 
	if( g>255 )	g=255;
	out.g = (u8)g;
	f32 b = interp3p(vtx1.b*vtx1.rhw,vtx2.b*vtx2.rhw,vtx3.b*vtx3.rhw,s1,s2)/out.rhw;
	if( b<0 )	b=0; 
	if( b>255 )	b=255;
	out.b = (u8)b;
	f32 a = interp3p(vtx1.a*vtx1.rhw,vtx2.a*vtx2.rhw,vtx3.a*vtx3.rhw,s1,s2)/out.rhw;
	if( a<0 )	a=0; 
	if( a>255 )	a=255;
	out.a = (u8)a;

	for( u32 i=0; i<SClipVertex::NUM_TEX_COORDS; i++ )
	{
		out.tcoord[i].u = interp3p(vtx1.tcoord[i].u*vtx1.rhw,vtx2.tcoord[i].u*vtx2.rhw,vtx3.tcoord[i].u*vtx3.rhw,s1,s2)/out.rhw;
		out.tcoord[i].v = interp3p(vtx1.tcoord[i].v*vtx1.rhw,vtx2.tcoord[i].v*vtx2.rhw,vtx3.tcoord[i].v*vtx3.rhw,s1,s2)/out.rhw;
	}

	if( out.rhw < 0 )	
	{
		return false;
	}
	else
		return true;
}

bool Clip1TriangleForNegW(const SClipVertex &vtx1, const SClipVertex &vtx2, const SClipVertex &vtx3, s32 &dstidx)
{
#ifdef DEBUG_POLYCLIPPER
	if( pauseAtNext && logTriangles ) 
	{
		TRACE0("To clip triangle for negative w:");
		TRACE4("Vtx 1: %f, %f, %f, %f", vtx1.x, vtx1.y, vtx1.z, 1/vtx1.rhw);
		TRACE4("Vtx 2: %f, %f, %f, %f", vtx2.x, vtx2.y, vtx2.z, 1/vtx2.rhw);
		TRACE4("Vtx 3: %f, %f, %f, %f", vtx3.x, vtx3.y, vtx3.z, 1/vtx3.rhw);
	}
#endif

	s32 lno = 0;
	if( vtx1.rhw >= 0 || vtx2.rhw >= 0 )	Create1LineEq(lines[lno++],vtx1,vtx2,vtx3);
	if( fabsf(EvaLine(lines[0],vtx3.x, vtx3.y) ) < 1 )	// vtx1, vtx2, vtx3 are linear
	{
		return false;
	}

	if( vtx1.rhw >= 0 || vtx3.rhw >= 0 )	Create1LineEq(lines[lno++],vtx1,vtx3,vtx2);
	if( vtx2.rhw >= 0 || vtx3.rhw >= 0 )	Create1LineEq(lines[lno++],vtx2,vtx3,vtx1);

	std::vector<v3> pts[2];

	v3 p;
	p.x = 0;
	p.y = 0;
	pts[0].push_back(p);
	p.x = gScreenWidth;
	p.y = 0;
	pts[0].push_back(p);
	p.x = gScreenWidth;
	p.y = gScreenHeight;
	pts[0].push_back(p);
	p.x = 0;
	p.y = gScreenHeight;
	pts[0].push_back(p);


	for( s32 j=0; j<lno; j++ )
	{
		ClipFor1LineXY(pts[j%2], pts[(j+1)%2], lines[j]);
		if( pts[(j+1)%2].size() < 3 )
			return false;
	}

	std::vector<v3> &ps = pts[lno%2];
	s32 size = ps.size();

	if( gZCompare )
	{
		for( s32 k=0; k<size; k++ )
		{
			Interp1PtZ(ps[k],vtx1,vtx2,vtx3);
		}

		ClipFor1LineZ(pts[lno%2], pts[(lno+1)%2], true);	// Clip for near plane
		if( pts[(lno+1)%2].size() < 3 )
			return false;
		ClipFor1LineZ(pts[(lno+1)%2], pts[lno%2], false);	// clip for far plane
		if( pts[lno%2].size() < 3 )
			return false;

		size = ps.size();
	}

	// now compute z and rhw values
	s32 i;
	SClipVertex  *newvertexes = new SClipVertex[size];

	for( i=0; i<size; i++ )
	{
		v3 &p = ps[i];
		if( fabsf(p.x-vtx1.x)<1 && fabsf(p.y-vtx1.y)<1 )
		{
			memcpy(&newvertexes[i], &vtx1, sizeof(SClipVertex) );
			continue;
		}
		else if( fabsf(p.x-vtx2.x)<1 && fabsf(p.y-vtx2.y)<1 )
		{
			memcpy(&newvertexes[i], &vtx2, sizeof(SClipVertex) );
			continue;
		}
		if( fabsf(p.x-vtx3.x)<1 && fabsf(p.y-vtx3.y)<1 )
		{
			memcpy(&newvertexes[i], &vtx3, sizeof(SClipVertex) );
			continue;
		}

		SClipVertex newvtx;
        if( Interp1Pt(ps[i],vtx1,vtx2,vtx3,newvtx) )
		{
			memcpy(&newvertexes[i], &newvtx, sizeof(SClipVertex) );
		}
		else
		{
			// negative w
			memcpy(&newvertexes[i], &newvtx, sizeof(SClipVertex) );
			//return false;	
		}
	}

#ifdef DEBUG_POLYCLIPPER
	if( pauseAtNext && logTriangles ) 
	{
		TRACE0("Clipped x and y:");
		for( i=0; i<size; i++ )
		{
			DebuggerAppendMsg("Vtx %d: %f, %f", i, ps[i].x, ps[i].y);
		}
		TRACE0("Generated new vertexes after clipping for W:");
		for( i=0; i<size; i++ )
		{
			DebuggerAppendMsg("Vtx %d: %f, %f, %f, %f", i, newvertexes[i].x, newvertexes[i].y, newvertexes[i].z, 1/newvertexes[i].rhw);
		}
	}
#endif

	for( i=0; i<size-2; i++ )
	{
		CopyVertexData(0,   newvertexes, dstidx++, g_clippedVtxBuffer);
		CopyVertexData(i+1, newvertexes, dstidx++, g_clippedVtxBuffer);
		CopyVertexData(i+2, newvertexes, dstidx++, g_clippedVtxBuffer);
	}

	delete [] newvertexes;

	return true;
}



bool Clip1TriangleForZ(const SClipVertex &vtx1, const SClipVertex &vtx2, const SClipVertex &vtx3, s32 &dstidx)
{
#ifdef DEBUG_POLYCLIPPER
	if( pauseAtNext && logTriangles ) 
	{
		TRACE0("To clip triangle for z:");
		TRACE4("Vtx 1: %f, %f, %f, %f", vtx1.x, vtx1.y, vtx1.z, 1/vtx1.rhw);
		TRACE4("Vtx 2: %f, %f, %f, %f", vtx2.x, vtx2.y, vtx2.z, 1/vtx2.rhw);
		TRACE4("Vtx 3: %f, %f, %f, %f", vtx3.x, vtx3.y, vtx3.z, 1/vtx3.rhw);
	}
#endif

	std::vector<v3> pts[2];

	v3 p;
	p.x = vtx1.x;
	p.y = vtx1.y;
	p.z = vtx1.z;
	pts[0].push_back(p);
	p.x = vtx2.x;
	p.y = vtx2.y;
	p.z = vtx2.z;
	pts[0].push_back(p);
	p.x = vtx3.x;
	p.y = vtx3.y;
	p.z = vtx3.z;
	pts[0].push_back(p);


	ClipFor1LineZ(pts[0], pts[1], true);	// Clip for near plane
	if( pts[1].size() < 3 )
		return false;
	ClipFor1LineZ(pts[1], pts[0], false);	// clip for far plane
	if( pts[0].size() < 3 )
		return false;

	std::vector<v3> &ps = pts[0];
	s32 size = ps.size();

	// now compute x, y and rhw values
	s32 i;
	SClipVertex  *newvertexes = new SClipVertex[size];

	for( i=0; i<size; i++ )
	{
		v3 &p = ps[i];
		if( fabsf(p.x-vtx1.x)<1 && fabsf(p.y-vtx1.y)<1 )
		{
			memcpy(&newvertexes[i], &vtx1, sizeof(SClipVertex) );
			continue;
		}
		else if( fabsf(p.x-vtx2.x)<1 && fabsf(p.y-vtx2.y)<1 )
		{
			memcpy(&newvertexes[i], &vtx2, sizeof(SClipVertex) );
			continue;
		}
		if( fabsf(p.x-vtx3.x)<1 && fabsf(p.y-vtx3.y)<1 )
		{
			memcpy(&newvertexes[i], &vtx3, sizeof(SClipVertex) );
			continue;
		}

		SClipVertex newvtx;
		Interp1Pt(ps[i],vtx1,vtx2,vtx3,newvtx);
		memcpy(&newvertexes[i], &newvtx, sizeof(SClipVertex) );
	}

#ifdef DEBUG_POLYCLIPPER
	if( pauseAtNext && logTriangles ) 
	{
		TRACE0("Clipped z:");
		for( i=0; i<size; i++ )
		{
			DebuggerAppendMsg("Vtx %d: %f, %f, %f", i, ps[i].x, ps[i].y, ps[i].z);
		}
		TRACE0("Generated new vertexes after clipping for z:");
		for( i=0; i<size; i++ )
		{
			DebuggerAppendMsg("Vtx %d: %f, %f, %f, %f", i, newvertexes[i].x, newvertexes[i].y, newvertexes[i].z, 1/newvertexes[i].rhw);
		}
	}
#endif

	for( i=0; i<size-2; i++ )
	{
		CopyVertexData(0,   newvertexes, dstidx++, g_clippedVtxBuffer);
		CopyVertexData(i+1, newvertexes, dstidx++, g_clippedVtxBuffer);
		CopyVertexData(i+2, newvertexes, dstidx++, g_clippedVtxBuffer);
	}

	delete [] newvertexes;

	return true;
}
