/*
Copyright (C) 2006 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "PspRomSelector.h"
#include "ColourPulser.h"
#include "UIContext.h"
#include "UIScreen.h"

#include <psptypes.h>
#include <pspkernel.h>
#include <pspctrl.h>
#include <pspdisplay.h>
#include <pspgu.h>

#include "PSPGraphics\Vector2.h"
#include "PSPGraphics\DrawText.h"
#include "PSPGraphics\PSPColour.h"

#include "Core\ROM.h"
#include "Interface\IniFile.h"
#include "Interface\RomDB.h"

#include "DaedIO.h"

#include <string>
#include <vector>
#include <functional>

namespace
{
	const u32 COLOUR_RED = 0xff0000ff;
	const u32 COLOUR_TEXT_WHITE = 0xffeeeeee;

	const char * const		BACKGROUND_FILENAME = "Resources/plain_background.png";

	const char * const		gRomsDirectories[] = 
	{
		"Roms/",
		"ms0:/n64/",
	};

	const u32				TEXT_AREA_TOP = 55;
	const u32				TEXT_AREA_LEFT = 40;
	const u32				TEXT_AREA_WIDTH = 400;
	const u32				TEXT_AREA_HEIGHT = 176;
	const u32				ROW_HEIGHT = 16;

	const u32				TITLE_AREA_TOP = 15;
	const u32				TITLE_AREA_LEFT = 0;
	const u32				TITLE_AREA_RIGHT = 480;

	const char * const		ROM_SELECTION_TEXT = "Rom Selection";

	const u32				MS_PER_COLOUR_CYCLE = 1200;

	const u32				COLOUR_SELECTION_DIM	= MAKE_COLOUR_U32( 0xe0, 0xe0, 0x80 );
	const u32				COLOUR_SELECTION_BRIGHT	= MAKE_COLOUR_U32( 0xff, 0xff, 0x80 );
}

//*************************************************************************************
//
//*************************************************************************************
struct SRomInfo
{
	std::string		mFilename;

	RomID			mRomID;
	u32				mRomSize;
	ECicType		mCicType;

	RomSettings		mSettings;

	SRomInfo( const char * filename )
		:	mFilename( filename )
	{
		memset( &mSettings, 0, sizeof( mSettings ) );

		if ( ROM_GetRomID( filename, &mRomID, &mRomSize, &mCicType ) )
		{
			strcpy( mSettings.szGameName, IO::Path::FindFileName( filename ) );
			if ( !CIniFile::Get()->GetSettings( mRomID, &mSettings ) )
			{
				// Create new entry, add
				ROM_ResetSettings( &mSettings );
				strcpy( mSettings.szComment, "No Comment" );

				//
				// We want to get the "internal" name for this rom from the header
				// Failing that, use the filename
				//
				if ( !ROM_GetRomName( filename, mSettings.szGameName ) )
				{
					memset( mSettings.szGameName, 0, sizeof( mSettings.szGameName ) );
					strncpy( mSettings.szGameName, IO::Path::FindFileName( filename ), sizeof( mSettings.szGameName ) );
				}
				CIniFile::Get()->SetSettings( mRomID, mSettings );
			}
		}
		else
		{
			strcpy( mSettings.szGameName, "Can't get rom info" );
		}

	}
};

//*************************************************************************************
//
//*************************************************************************************
static bool SortByGameName( const SRomInfo * a, const SRomInfo * b )
{
	return stricmp( a->mSettings.szGameName, b->mSettings.szGameName ) < 0;
}



//*************************************************************************************
//
//*************************************************************************************
class IPspRomSelector : public CPspRomSelector, CUIScreen
{
		typedef std::vector<SRomInfo*>	RomInfoList;
	public:

		IPspRomSelector();
		~IPspRomSelector();

		// CPspRomSelector
		virtual void				Run( CUIContext * p_context );
		virtual void				GetSelectedRom( char * p_filename );

		// CUIScreen
		virtual void				Update( float elapsed_time, const v2 & stick, u32 old_buttons, u32 new_buttons );
		virtual void				Render();
		virtual bool				IsFinished() const									{ return mRomSelected; }

	private:
				void				AddRomDirectory(const char * p_roms_dir, RomInfoList & roms);

	private:
		RomInfoList					mRomsList;
		u32							mCurrentSelection;
		s32							mCurrentScrollOffset;
		float						mSelectionAccumulator;
		std::string					mSelectedRom;
		bool						mRomSelected;

		CColourPulser				mColourPulser;
};

// XXXX If this is added as a member var seem to get random bus errors!!
SceIoDirent					gDirEntry;

// XXXX Need to reevaluate whether this safety buffer is needed anymore..
static u8					gBytes[128];


//*************************************************************************************
//
//*************************************************************************************
CPspRomSelector::~CPspRomSelector()
{
}

//*************************************************************************************
//
//*************************************************************************************
CPspRomSelector *	CPspRomSelector::Create()
{
	return new IPspRomSelector;
}

//*************************************************************************************
//
//*************************************************************************************
IPspRomSelector::IPspRomSelector()
:	mCurrentSelection( 0 )
,	mCurrentScrollOffset( 0 )
,	mSelectionAccumulator( 0 )
,	mRomSelected( false )
,	mColourPulser( COLOUR_SELECTION_DIM, COLOUR_SELECTION_BRIGHT, MS_PER_COLOUR_CYCLE )
{
	use(gBytes);

	for( u32 i = 0; i < ARRAYSIZE( gRomsDirectories ); ++i )
	{
		AddRomDirectory( gRomsDirectories[ i ], mRomsList );
	}

	sort( mRomsList.begin(), mRomsList.end(), SortByGameName );
}

//*************************************************************************************
//
//*************************************************************************************
IPspRomSelector::~IPspRomSelector()
{
	CRomDB::Get()->Reset();

	for(RomInfoList::iterator it = mRomsList.begin(); it != mRomsList.end(); ++it)
	{
		SRomInfo *	p_rominfo( *it );

		delete p_rominfo;
	}
	mRomsList.clear();
}

//*************************************************************************************
//
//*************************************************************************************
bool	CFileSystem_FindFirstFile( const char * path, SceUID & handle )
{
	handle = sceIoDopen( path );

	return ( handle >= 0 );
}

//*************************************************************************************
//
//*************************************************************************************
bool	CFileSystem_FindNextFile( SceIoDirent & dir_entry, SceUID handle )
{
	DAEDALUS_ASSERT( handle >= 0, "Cannot search with invalid directory handle" );

	const s32	error_code( sceIoDread( handle, &dir_entry ) );

	DAEDALUS_ASSERT( error_code >= 0, "Error reading directory entry" );

	if ( error_code > 0 )
	{
		return true;
	}

	return false;
}

//*************************************************************************************
//
//*************************************************************************************
bool	CFileSystem_FindCloseFile( SceUID handle )
{
	DAEDALUS_ASSERT( handle >= 0, "Trying to close an invalid directory handle" );

	return ( sceIoDclose( handle ) >= 0 );
}

//*************************************************************************************
//
//*************************************************************************************
void	IPspRomSelector::AddRomDirectory(const char * p_roms_dir, RomInfoList & roms)
{
	SceUID				dir_handle;
	if(CFileSystem_FindFirstFile( p_roms_dir, dir_handle ))
	{
		while(CFileSystem_FindNextFile( gDirEntry, dir_handle ))
		{
			const char * rom_filename( gDirEntry.d_name );
			const char * last_period( strrchr( rom_filename, '.' ) );
			if(last_period != NULL)
			{
				if(stricmp(last_period, ".v64") == 0 ||
				   stricmp(last_period, ".z64") == 0 ||
				   stricmp(last_period, ".rom") == 0 ||
				   stricmp(last_period, ".zip") == 0)
				{
					std::string		full_path;

					full_path = p_roms_dir;
					full_path += rom_filename;

					SRomInfo *	p_rom_info = new SRomInfo( full_path.c_str() );

					roms.push_back( p_rom_info );
				}
			}
		}

		CFileSystem_FindCloseFile( dir_handle );
	}
}

//*************************************************************************************
//
//*************************************************************************************
void IPspRomSelector::Render()
{
	s32		x,y;

	x = DrawTextUtilities::AlignText( TITLE_AREA_LEFT, TITLE_AREA_RIGHT, ROM_SELECTION_TEXT, DrawTextUtilities::AT_CENTRE );
	y = TITLE_AREA_TOP;

	DrawTextUtilities::DrawShadowText( x, y, ROM_SELECTION_TEXT, COLOUR_RED );


	x = TEXT_AREA_LEFT;
	y = TEXT_AREA_TOP + mCurrentScrollOffset;

	sceGuEnable(GU_SCISSOR_TEST);
	sceGuScissor(TEXT_AREA_LEFT, TEXT_AREA_TOP, TEXT_AREA_LEFT+TEXT_AREA_WIDTH, TEXT_AREA_TOP+TEXT_AREA_HEIGHT);

	for(u32 i = 0; i < mRomsList.size(); ++i)
	{
		const char * p_gamename( mRomsList[ i ]->mSettings.szGameName );
		u32		colour;

		//
		// Check if this entry would be onscreen
		//
		if((y+ROW_HEIGHT) >= s32(TEXT_AREA_TOP) && y < s32(TEXT_AREA_TOP + TEXT_AREA_HEIGHT))
		{
			if(i == mCurrentSelection)
			{
				colour = mColourPulser.GetCurrentColour();
			}
			else
			{
				colour = COLOUR_TEXT_WHITE;
			}

			DrawTextUtilities::DrawShadowText( x, y, p_gamename, colour );
		}
		y += 16;
	}

	// Restore scissoring
	sceGuScissor(0,0, 480,272);
}

//*************************************************************************************
//
//*************************************************************************************
void	IPspRomSelector::Update( float elapsed_time, const v2 & stick, u32 old_buttons, u32 new_buttons )
{
	static const float	SCROLL_RATE_PER_SECOND = 15.0f;		// 5 roms/second
	static const float	MAX_SCROLL_PER_FRAME = 2.0f;

	// Simple rom chooser
	if(new_buttons & PSP_CTRL_LTRIGGER)
	{
		mSelectionAccumulator += -SCROLL_RATE_PER_SECOND * elapsed_time;
	}
	if(new_buttons & PSP_CTRL_RTRIGGER)
	{
		mSelectionAccumulator += +SCROLL_RATE_PER_SECOND * elapsed_time;
	}

	mSelectionAccumulator += stick.y * SCROLL_RATE_PER_SECOND * elapsed_time; 

	if(old_buttons != new_buttons)
	{
		if(new_buttons & PSP_CTRL_UP)
		{
			if(mCurrentSelection > 0)
			{
				mCurrentSelection--;
			}
		}
		if(new_buttons & PSP_CTRL_DOWN)
		{
			if(mCurrentSelection < mRomsList.size() - 1)
			{
				mCurrentSelection++;
			}
		}

		if((new_buttons & PSP_CTRL_START) ||
			(new_buttons & PSP_CTRL_CROSS))
		{
			mRomSelected = true;
			if(mCurrentSelection < mRomsList.size())
			{
				mSelectedRom = mRomsList[ mCurrentSelection ]->mFilename.c_str();
			}
		}
	}

	//
	//	Clamp the accumulator
	//
	if(mSelectionAccumulator > MAX_SCROLL_PER_FRAME)
	{
		mSelectionAccumulator = MAX_SCROLL_PER_FRAME;
	}
	else if(mSelectionAccumulator < -MAX_SCROLL_PER_FRAME)
	{
		mSelectionAccumulator = -MAX_SCROLL_PER_FRAME;
	}

	//
	//	Apply the selection accumulator
	//
	while(mSelectionAccumulator >= 1.0f)
	{
		if(mCurrentSelection < mRomsList.size() - 1)
		{
			mCurrentSelection++;
		}
		mSelectionAccumulator -= 1.0f;
	}
	while(mSelectionAccumulator <= -1.0f)
	{
		if(mCurrentSelection > 0)
		{
			mCurrentSelection--;
		}
		mSelectionAccumulator += 1.0f;
	}

	//
	//	Scroll to keep things in view
	//
	s32		current_selection_y = (mCurrentSelection * ROW_HEIGHT) + (ROW_HEIGHT/2) + mCurrentScrollOffset;

	s32		adjust_amount( (TEXT_AREA_HEIGHT/2) - current_selection_y );

	float d( 1.0f - powf(0.995f, elapsed_time * 1000.0f) );

	mCurrentScrollOffset += s32(float(adjust_amount) * d);

	s32		elapsed_ms = s32( 1000.0f * elapsed_time );

	mColourPulser.Update( elapsed_ms );
}

//*************************************************************************************
//
//*************************************************************************************
void	IPspRomSelector::Run( CUIContext * p_context )
{
	p_context->SetBackgroundTexture( BACKGROUND_FILENAME );
	p_context->Run( this );
}

//*************************************************************************************
//
//*************************************************************************************
void	IPspRomSelector::GetSelectedRom( char * p_filename )
{
	if( mRomSelected )
	{
		strcpy( p_filename, mSelectedRom.c_str() );
	}
	else
	{
		strcpy( p_filename, "" );
	}
}
