/*
Copyright (C) 2006 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "PspRomSettings.h"

#include "UIContext.h"
#include "UIScreen.h"
#include "ColourPulser.h"
#include "PSPGraphics\DrawText.h"
#include "PSPGraphics\PSPColour.h"

#include "Core\ROM.h"

#include <pspctrl.h>
#include <pspgu.h>
namespace
{
	const u32		TITLE_AREA_TOP = 15;
	const u32		TITLE_AREA_LEFT = 16;
	const u32		TITLE_AREA_RIGHT = 480-16;

	const u32		TEXT_AREA_TOP = 15+16+16;
	const u32		TEXT_AREA_LEFT = 40;
	const u32		TEXT_AREA_RIGHT = 440;

	const u32		DESCRIPTION_AREA_TOP = 272-32-10;		// 2 rows
	const u32		DESCRIPTION_AREA_LEFT = 16;
	const u32		DESCRIPTION_AREA_RIGHT = 480-16;

	const char * const ROM_SETTINGS_TEXT = "Rom Settings";

	const u32		MS_PER_COLOUR_CYCLE = 1200;

	const u32		COLOUR_SELECTION_DIM	= MAKE_COLOUR_U32( 0xe0, 0xe0, 0x80 );
	const u32		COLOUR_SELECTION_BRIGHT	= MAKE_COLOUR_U32( 0xff, 0xff, 0x80 );

	const u32		COLOUR_TEXT_WHITE = 0xffeeeeee;
	const u32		COLOUR_TEXT_WHITE_DISABLED = 0xffd0d0d0;
	const u32		COLOUR_TEXT_RED = 0xff0000ff;
	const u32		COLOUR_TEXT_RED_DISABLED = 0xffb0b0ff;

	const char * const	BACKGROUND_FILENAME = "Resources/plain_background.png";

	enum	ERomSettingType
	{
		RS_SAVE_TYPE = 0,
		RS_EXPANSION_PAK,
		RS_DYNAREC,

		NUM_ROM_SETTINGS,
	};

	const ERomSettingType	RS_FIRST_SETTING = RS_SAVE_TYPE;
	const ERomSettingType	RS_LAST_SETTING = RS_DYNAREC;

	struct SRomSettingInfo
	{
		const char *		Setting;
		const char *		Description;
	};

	const	SRomSettingInfo		gRomSettingInfo[] = 
	{
		{ "Save Type", "The save game mechanism used for this rom (this setting is read-only)" },
		{ "Expansion Pak", "Whether the expansion pak used or required by this rom (not required for many roms)" },
		{ "Dynamic Recompilation", "Whether dynamic recompilation is enabled for this rom" },
	};

}

//*************************************************************************************
//
//*************************************************************************************
class IPspRomSettings : public CPspRomSettings, public CUIScreen
{
	public:

		IPspRomSettings();
		~IPspRomSettings();

		// CPspRomSelettings
		virtual EScreenStatus		Run( CUIContext * p_context );
		virtual void				SetSettings( const RomSettings * p_settings );
		virtual void				GetSettings( RomSettings * p_settings );

		// CUIScreen
		virtual void				Update( float elapsed_time, const v2 & stick, u32 old_buttons, u32 new_buttons );
		virtual void				Render();
		virtual bool				IsFinished() const									{ return mIsFinished; }

				bool				IsSettingReadOnly( ERomSettingType rom_setting ) const;
				void				NextSetting( ERomSettingType rom_setting );
				void				PreviousSetting( ERomSettingType rom_setting );
				const char *		GetSettingName( ERomSettingType rom_setting ) const ;

	private:
		EScreenStatus				mScreenStatus;
		bool						mIsFinished;
		ERomSettingType				mSelectedSetting;

		CColourPulser				mColourPulser;

		RomSettings					mRomSettings;
};

//*************************************************************************************
//
//*************************************************************************************
CPspRomSettings::~CPspRomSettings()
{
}

//*************************************************************************************
//
//*************************************************************************************
CPspRomSettings *	CPspRomSettings::Create()
{
	return new IPspRomSettings;
}

//*************************************************************************************
//
//*************************************************************************************
IPspRomSettings::IPspRomSettings()
:	mScreenStatus( SS_OK )
,	mIsFinished( false )
,	mSelectedSetting( RS_FIRST_SETTING )
,	mColourPulser( COLOUR_SELECTION_DIM, COLOUR_SELECTION_BRIGHT, MS_PER_COLOUR_CYCLE )
{
}

//*************************************************************************************
//
//*************************************************************************************
IPspRomSettings::~IPspRomSettings()
{
}

//*************************************************************************************
//
//*************************************************************************************
void	IPspRomSettings::Update( float elapsed_time, const v2 & stick, u32 old_buttons, u32 new_buttons )
{
	if(old_buttons != new_buttons)
	{
		if(new_buttons & PSP_CTRL_START)
		{
			mIsFinished = true;
			mScreenStatus = SS_OK;
		}
		if(new_buttons & PSP_CTRL_SELECT)
		{
			mIsFinished = true;
			mScreenStatus = SS_BACK;
		}

		if( new_buttons & PSP_CTRL_UP )
		{
			if( mSelectedSetting > RS_FIRST_SETTING )
			{
				mSelectedSetting = ERomSettingType( mSelectedSetting - 1 );
			}
		}
		if( new_buttons & PSP_CTRL_DOWN )
		{
			if( mSelectedSetting < RS_LAST_SETTING )
			{
				mSelectedSetting = ERomSettingType( mSelectedSetting + 1 );
			}
		}

		if( new_buttons & PSP_CTRL_LEFT )
		{
			if( !IsSettingReadOnly( mSelectedSetting ) )
			{
				PreviousSetting( mSelectedSetting );
			}
		}
		if( new_buttons & PSP_CTRL_RIGHT )
		{
			if( !IsSettingReadOnly( mSelectedSetting ) )
			{
				NextSetting( mSelectedSetting );
			}
		}
	}

	s32		elapsed_ms = s32( 1000.0f * elapsed_time );
	mColourPulser.Update( elapsed_ms );
}

//*************************************************************************************
//
//*************************************************************************************
void	IPspRomSettings::Render()
{
	s32		x,y;

	x = DrawTextUtilities::AlignText( TITLE_AREA_LEFT, TITLE_AREA_RIGHT, ROM_SETTINGS_TEXT, DrawTextUtilities::AT_CENTRE );
	y = TITLE_AREA_TOP;

	DrawTextUtilities::DrawShadowText( x, y, ROM_SETTINGS_TEXT, COLOUR_TEXT_RED );

	y = TEXT_AREA_TOP + 0;

	const u32	font_height( CDrawText::GetFontHeight() );

	for( u32 i = 0; i < NUM_ROM_SETTINGS; ++i)
	{
		ERomSettingType		setting = ERomSettingType( i );

		bool				read_only( IsSettingReadOnly( setting ) );

		u32		colour;
		if( setting == mSelectedSetting )
		{
			colour = read_only ? COLOUR_TEXT_RED_DISABLED : mColourPulser.GetCurrentColour();
		}
		else
		{
			colour = read_only ? COLOUR_TEXT_WHITE_DISABLED : COLOUR_TEXT_WHITE;
		}

		x = TEXT_AREA_LEFT;
		DrawTextUtilities::DrawShadowText( x, y, gRomSettingInfo[i].Setting, colour );

		const char * p_setting( GetSettingName( setting ) );
		x = DrawTextUtilities::AlignText( TEXT_AREA_LEFT, TEXT_AREA_RIGHT, p_setting, DrawTextUtilities::AT_RIGHT );
		DrawTextUtilities::DrawShadowText( x, y, p_setting, colour );
		y += font_height;
	}

	x = DESCRIPTION_AREA_LEFT;
	y = DESCRIPTION_AREA_TOP;

	const char *		p_description( gRomSettingInfo[mSelectedSetting].Description );
	std::vector<u32>	lengths;
	DrawTextUtilities::WrapText( DESCRIPTION_AREA_RIGHT - DESCRIPTION_AREA_LEFT, p_description, strlen( p_description ), lengths );
	for( u32 i = 0; i < lengths.size(); ++i )
	{
		DrawTextUtilities::DrawShadowText( x, y, p_description, lengths[ i ], COLOUR_TEXT_WHITE );
		p_description += lengths[ i ];
		y += font_height;
	}
}

//*************************************************************************************
//
//*************************************************************************************
CPspRomSettings::EScreenStatus	IPspRomSettings::Run( CUIContext * p_context )
{
	p_context->SetBackgroundTexture( BACKGROUND_FILENAME );
	p_context->Run( this );

	return mScreenStatus;
}

//*************************************************************************************
//
//*************************************************************************************
void	IPspRomSettings::SetSettings( const RomSettings * p_settings )
{
	mRomSettings = *p_settings;
}

//*************************************************************************************
//
//*************************************************************************************
void	IPspRomSettings::GetSettings( RomSettings * p_settings )
{
	*p_settings = mRomSettings;
}

//*************************************************************************************
//
//*************************************************************************************
bool	IPspRomSettings::IsSettingReadOnly( ERomSettingType rom_setting ) const
{
	switch( rom_setting )
	{
	case RS_SAVE_TYPE:
		return true;
	case RS_EXPANSION_PAK:
		return false;
	case RS_DYNAREC:
		return false;

	default:
		DAEDALUS_ERROR( "Unhandled rom setting" );
		return false;
	}

}

//*************************************************************************************
//
//*************************************************************************************
void	IPspRomSettings::NextSetting( ERomSettingType rom_setting )
{
	switch( rom_setting )
	{
	case RS_SAVE_TYPE:
		if(mRomSettings.SaveType < NUM_SAVE_TYPES-1)
			mRomSettings.SaveType = ESaveType( mRomSettings.SaveType + 1 );
		else
			mRomSettings.SaveType = SAVE_TYPE_UNKNOWN;
		break;

	case RS_EXPANSION_PAK:
		mRomSettings.ExpansionPakEnabled = !mRomSettings.ExpansionPakEnabled;
		break;

	case RS_DYNAREC:
		mRomSettings.DynarecEnabled = !mRomSettings.DynarecEnabled;
		break;

	default:
		DAEDALUS_ERROR( "Unhandled rom setting" );
		break;
	}
}

//*************************************************************************************
//
//*************************************************************************************
void	IPspRomSettings::PreviousSetting( ERomSettingType rom_setting )
{
	switch( rom_setting )
	{
	case RS_SAVE_TYPE:
		if(mRomSettings.SaveType > SAVE_TYPE_UNKNOWN)
			mRomSettings.SaveType = ESaveType( mRomSettings.SaveType - 1 );
		else
			mRomSettings.SaveType = ESaveType( NUM_SAVE_TYPES-1 );
		break;

	case RS_EXPANSION_PAK:
		mRomSettings.ExpansionPakEnabled = !mRomSettings.ExpansionPakEnabled;
		break;

	case RS_DYNAREC:
		mRomSettings.DynarecEnabled = !mRomSettings.DynarecEnabled;
		break;

	default:
		DAEDALUS_ERROR( "Unhandled rom setting" );
		break;
	}
}

//*************************************************************************************
//
//*************************************************************************************
const char *	IPspRomSettings::GetSettingName( ERomSettingType rom_setting ) const 
{
	switch( rom_setting )
	{
	case RS_SAVE_TYPE:
		return mRomSettings.SaveType == SAVE_TYPE_UNKNOWN ? "None" : ROM_GetSaveTypeName( mRomSettings.SaveType );

	case RS_EXPANSION_PAK:
		return mRomSettings.ExpansionPakEnabled ? "Yes" : "No";

	case RS_DYNAREC:
		return mRomSettings.DynarecEnabled ? "Enabled" : "Disabled";

	default:
		DAEDALUS_ERROR( "Unhandled rom setting" );
		return "?";
	}
}

