/*
Copyright (C) 2006 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "UIContext.h"

#include <psptypes.h>
#include <pspkernel.h>
#include <pspdebug.h>
#include <pspctrl.h>
#include <pspdisplay.h>
#include <pspgu.h>
#include <pspgum.h>

#include <functional>

#include "UIScreen.h"

#include "PSPGraphics\FFXMaths.h"
#include "PSPGraphics\Vector2.h"
#include "PSPGraphics\Vector3.h"
#include "PSPGraphics\GraphicsContext.h"
#include "PSPGraphics\DrawText.h"
#include "PSPGraphics\NativeTexture.h"

#define GL_TRUE                           1
#define GL_FALSE                          0

static const char * const		BACKGROUND_FILENAME = "Resources/plain_background.png";
static const u32				BACKGROUND_WIDTH = 480;
static const u32				BACKGROUND_HEIGHT = 272;

struct BackgroundVtx
{
    v2	t0;
    v3	pos;
};

DAEDALUS_STATIC_ASSERT( sizeof(BackgroundVtx) == 20 );

#define BACKGROUND_VERTEX_FLAGS (GU_TEXTURE_32BITF|GU_VERTEX_32BITF )

// XXXX
extern float	GetElapsedTime();
extern void		ResetElapsedTime();

//*************************************************************************************
//
//*************************************************************************************
class IUIContext : public CUIContext
{
	public:

		IUIContext();
		~IUIContext();

		virtual void				Run( CUIScreen * p_screen );
		virtual void				SetBackgroundTexture( const char * p_filename );

	private:
				void				Render( CUIScreen * p_screen );

				void				RenderBackground();

	private:
		CNativeTexture *			mpBackgroundTexture;
};

//*************************************************************************************
//
//*************************************************************************************
CUIContext::~CUIContext()
{
}

//*************************************************************************************
//
//*************************************************************************************
CUIContext *	CUIContext::Create()
{
	return new IUIContext;
}

//*************************************************************************************
//
//*************************************************************************************
IUIContext::IUIContext()
:	mpBackgroundTexture( CNativeTexture::Create( BACKGROUND_WIDTH, BACKGROUND_HEIGHT ) )
{
	mpBackgroundTexture->LoadPng( BACKGROUND_FILENAME );
}

//*************************************************************************************
//
//*************************************************************************************
IUIContext::~IUIContext()
{
	SAFE_RELEASE( mpBackgroundTexture );

	// Clear everything to black
	for( u32 i = 0; i < 2; ++i )
	{
		CGraphicsContext::Get()->BeginFrame();
		CGraphicsContext::Get()->Clear( true, true );
		CGraphicsContext::Get()->EndFrame();
		CGraphicsContext::Get()->UpdateFrame( false );
	}
}


//*************************************************************************************
//
//*************************************************************************************
void	IUIContext::SetBackgroundTexture( const char * p_filename )
{
	mpBackgroundTexture->LoadPng( p_filename );
}

//*************************************************************************************
//
//*************************************************************************************
void	IUIContext::RenderBackground()
{
	u32				num_verts( 2 );
	BackgroundVtx*	p_verts = (BackgroundVtx*)sceGuGetMemory(num_verts*sizeof(BackgroundVtx));

	sceGuDisable(GU_DEPTH_TEST);
	sceGuDepthMask( GL_TRUE );	// GL_TRUE to disable z-writes
	sceGuShadeModel( GU_FLAT );

	sceGuTexFilter(GU_NEAREST,GU_NEAREST);
	sceGuTexScale(1.0f,1.0f);
	sceGuTexOffset(0.0f,0.0f);

	if(mpBackgroundTexture != NULL)
	{
		mpBackgroundTexture->InstallTexture();
	}

	sceGuDisable(GU_ALPHA_TEST);
	sceGuDisable(GU_BLEND);
	sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);

	sceGumMatrixMode(GU_PROJECTION);
	sceGumLoadIdentity();
	//sceGumOrtho(0, 480 +1, 272 +1, 0, 0.0f, -1.0f);

	v2		tex_uv0( (float)0, (float)0 );
	v2		tex_uv1( (float)BACKGROUND_WIDTH, (float)BACKGROUND_HEIGHT );

	p_verts[0].pos = v3( 0.0f, 0.0f, 0.0f );
	p_verts[0].t0 = v2( 0.0f, 0.0f );

	p_verts[1].pos = v3( (float)BACKGROUND_WIDTH, (float)BACKGROUND_HEIGHT, 0.0f );
	p_verts[1].t0 = v2( (float)BACKGROUND_WIDTH, (float)BACKGROUND_HEIGHT );

	sceGumDrawArray(GU_SPRITES,BACKGROUND_VERTEX_FLAGS|GU_TRANSFORM_2D,num_verts,NULL,p_verts);
}

//*************************************************************************************
//
//*************************************************************************************
void	IUIContext::Render( CUIScreen * p_screen )
{
	DAEDALUS_ASSERT( p_screen != NULL, "No screen supplied" );

	CGraphicsContext::Get()->BeginFrame();
	//CGraphicsContext::Get()->Clear( 0xff39A2F9, 0 );

	RenderBackground();

	p_screen->Render();

	CGraphicsContext::Get()->EndFrame();

	CGraphicsContext::Get()->UpdateFrame( true );
}

//*************************************************************************************
//
//*************************************************************************************
void	IUIContext::Run( CUIScreen * p_screen )
{
	DAEDALUS_ASSERT( p_screen != NULL, "No screen supplied" );

	SceCtrlData		pad;
	sceCtrlReadBufferPositive(&pad, 1);
	u32		oldButtons = pad.Buttons;

	static const s32	STICK_DEADZONE = 20;

	ResetElapsedTime();

	// Simple rom chooser
	while( !p_screen->IsFinished() )
	{
		float		elapsed_time( GetElapsedTime() );

		oldButtons = pad.Buttons;
		sceCtrlReadBufferPositive(&pad, 1);

		s32		stick_x( pad.Lx - 128 );
		s32		stick_y( pad.Ly - 128 );

		if(stick_x >= -STICK_DEADZONE && stick_x <= STICK_DEADZONE)
		{
			stick_x = 0;
		}
		if(stick_y >= -STICK_DEADZONE && stick_y <= STICK_DEADZONE)
		{
			stick_y = 0;
		}

		v2	stick;
		stick.x = float(stick_x) / 128.0f;
		stick.y = float(stick_y) / 128.0f;

		p_screen->Update( elapsed_time, stick, oldButtons, pad.Buttons );
		
		Render( p_screen );
	}
}
